/*

dirt.c - Directory Traversal - 16 bit OS/2 version

*/

/*...sincludes:0:*/
#define	INCL_BASE
#include <os2.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define	DIRT_C
#include "dirt.h"

/*...vdirt\46\h:0:*/
/*...e*/

typedef struct
	{
	HDIR hdir;
	USHORT cSearch;
	FILEFINDBUF ffb;
	char filename[CCHMAXPATHCOMP+1];
	BOOL empty, first;
	} DIRT;

/*...sdirt_open:0:*/
DIRT *dirt_open(const char *dirname, int *rc)
	{
	DIRT *dirt;
	char pattern[CCHMAXPATHCOMP+2+1];
	USHORT drc;
	int len = strlen(dirname);
	*rc = DIRTE_OK;
	if ( len == 0 || len > CCHMAXPATHCOMP )
		{
		*rc = DIRTE_NOT_FOUND;
		return NULL;
		}
	if ( (dirt = (DIRT *) malloc(sizeof(DIRT))) == NULL )
		{
		*rc = DIRTE_NO_MEMORY;
		return NULL;
		}
	memcpy(pattern, dirname, len);
	if ( pattern[len-1] == '/' || pattern[len-1] == '\\' )
		strcpy(pattern+len-1, "\\*");
	else if ( len == 2 && isalpha(pattern[0]) && pattern[1] == ':' )
		strcpy(pattern+len, "*");
	else
		strcpy(pattern+len, "\\*");
	dirt->hdir = HDIR_CREATE;
	dirt->cSearch = 1;
	drc = DosFindFirst(pattern, &(dirt->hdir), FILE_ARCHIVED|FILE_DIRECTORY|FILE_HIDDEN|FILE_SYSTEM|FILE_READONLY, &(dirt->ffb), sizeof(dirt->ffb), &(dirt->cSearch), 0L);
	if ( drc != 0 )
		{
		switch ( drc )
			{
			case ERROR_NO_MORE_FILES:
				dirt->empty = TRUE;
				return dirt;
			case ERROR_FILE_NOT_FOUND:
			case ERROR_PATH_NOT_FOUND:
				*rc = DIRTE_NOT_FOUND;
				break;
			default:
				*rc = DIRTE_GEN_ERROR;
				break;
			}
		free(dirt);
		return NULL;
		}
	dirt->empty = FALSE;
	dirt->first = TRUE;
	return dirt;
	}
/*...e*/
/*...sdirt_next:0:*/
const char *dirt_next(DIRT *dirt)
	{
	if ( dirt->empty )
		return NULL;
	else if ( dirt->first )
		dirt->first = FALSE;
	else
		{
		if ( DosFindNext(dirt->hdir, &(dirt->ffb), sizeof(dirt->ffb), &(dirt->cSearch)) != 0 )
			return NULL;
		}
	
	strncpy(dirt->filename, dirt->ffb.achName, dirt->ffb.cchName+1);
	dirt->filename[CCHMAXPATHCOMP] = '\0';
	return dirt->filename;
	}
/*...e*/
/*...sdirt_close:0:*/
void dirt_close(DIRT *dirt)
	{
	DosFindClose(dirt->hdir);
	free(dirt);
	}
/*...e*/

/*...sdirt_error:0:*/
const char *dirt_error(int rc)
	{
	const char *errlist[] =
		{
		NULL,
		"general error",
		"out of memory",
		"not found",
		};
	return errlist[rc];
	}
/*...e*/
