;
; crit.asm - Intercept Critical Error
;

;...smodel stuff:0:
		IFDEF	memS
		.MODEL	SMALL,C
sizec		EQU	0
sized		EQU	0
		ENDIF
		IFDEF	memC
		.MODEL	COMPACT,C
sizec		EQU	0
sized		EQU	1
		ENDIF
		IFDEF	memM
		.MODEL	MEDIUM,C
sizec		EQU	1
sized		EQU	0
		ENDIF
		IFDEF	memL
		.MODEL	LARGE,C
sizec		EQU	1
sized		EQU	1
		ENDIF
;...e
;...sequs:0:
crit		EQU	024H		; Critical error vector
dos		EQU	021H		; Dos BIOS interrupt
dos_set_ivec	EQU	025H		; Set interrupt vector
dos_get_ivec	EQU	035H		; Get interrupt vector

DOSCALL		MACRO	call_no
		MOV	AH,call_no
		INT	dos
		ENDM
;...e

		.CODE

;...svars:0:
inited		DB	0

old_crit	LABEL	DWORD
old_off_crit	DW	?
old_seg_crit	DW	?
;...e

;...svoid cdecl crit_init\40\void\41\:0:
crit_init	PROC	USES BX DX DS ES

		MOV	AL,crit
		DOSCALL	dos_get_ivec	; Get interrupt vector to ES:BX
		MOV	CS:[old_off_crit],BX
		MOV	CS:[old_seg_crit],ES

		MOV	AX,CS
		MOV	DS,AX
		MOV	DX,OFFSET _new_crit
		MOV	AL,crit
		DOSCALL	dos_set_ivec

		MOV	CS:[inited],1

		RET

crit_init	ENDP
;...e
;...svoid cdecl crit_deinit\40\void\41\:0:
crit_deinit	PROC	USES DX DS

		MOV	AL,CS:[inited]
		AND	AL,AL
		JZ	SHORT not_inited

		LDS	DX,CS:[old_crit]
		MOV	AL,crit
		DOSCALL	dos_set_ivec	; restore old critical error handler

not_inited:	RET

crit_deinit	ENDP
;...e

;...s_new_crit:0:
;
; New critical error handler interrupt server
;
; On entry:	Bit 7,AH = 0 => Disk related
;				Bit 0,AH = 0 => read
;					 = 1 => write
;				Bits 2 and 1 = 00 => DOS system files
;					     = 01 => FAT
;			 = 1 => Non-disk or possibly disk FAT
;		BP:SI -> Device Header Control Block
;
; On exit:	AL = 0 => Ignore error
;		   = 1 => Retry operation
;		   = 2 => Abort program using Int-23h (Ouch!)
;
_new_crit	PROC	FAR

		AND	AH,080H		; is it disk related ?
		JNZ	SHORT not_disk
		MOV	AL,0
		IRET

not_disk:	JMP	CS:[old_crit]	; chain out

_new_crit	ENDP
;...e

		END
