;
; key.asm - BIOS keyboard reader for use by raw keyboard readers
;
; Microsoft "C" 5.0 is crap since it will not return the zero flag
; from its int86() calls. Furthermore, the _bios_keybrd(_KEYBRD_READY)
; routines can loop indefinately if Ctrl-Break is pressed since its
; keycode is 0 and 0 is returned if no key ready. Thus a loop of the
; form :-
;
;	while ( !_bios_keybrd(_KEYBRD_READY) )
;		;
;
; will loop forever since the Ctrl-Break will never be removed from the
; buffer.
;
; This code will swallow all Ctrl-Breaks, the calling program will never get one
;

;...smodel stuff:0:
		IFDEF	memS
		.MODEL	SMALL,C
sizec		EQU	0
sized		EQU	0
		ENDIF
		IFDEF	memC
		.MODEL	COMPACT,C
sizec		EQU	0
sized		EQU	1
		ENDIF
		IFDEF	memM
		.MODEL	MEDIUM,C
sizec		EQU	1
sized		EQU	0
		ENDIF
		IFDEF	memL
		.MODEL	LARGE,C
sizec		EQU	1
sized		EQU	1
		ENDIF
;...e
;...sequs:0:
b_brk		EQU	01BH		; break vector used by BIOS
d_brk		EQU	023H		; break vector used by DOS
dos		EQU	021H		; DOS BIOS interrupt
dos_set_ivec	EQU	025H		; Set interrupt vector
dos_ctrl_break	EQU	033H		; Get/set ctrl-break status
dos_get_ivec	EQU	035H		; Get interrupt vector
key		EQU	016H		; Keyboard BIOS interrupt
key_read_ch	EQU	000H		; Wait and get character
key_ready	EQU	001H		; Test keyboard status
key_get_shift	EQU	002H		; Get shift status
key_read_ch_x	EQU	010H		; Extended wait and get character
key_ready_x	EQU	011H		; Extended test keyboard status
key_get_shift_x	EQU	012H		; Get extended shift status

DOSCALL		MACRO	call_no
		MOV	AH,call_no
		INT	dos
		ENDM

KEYCALL		MACRO	call_no
		MOV	AH,call_no
		INT	key
		ENDM
;...e

		.CODE

;...svars:0:
old_b_brk	LABEL	DWORD
old_off_b_brk	DW	?
old_seg_b_brk	DW	?

old_d_brk	LABEL	DWORD
old_off_d_brk	DW	?
old_seg_d_brk	DW	?

old_ctrl_break	DB	0

inited		DB	0

break_flag	DW	0

extended	DB	0		; If set NZ then BIOS 102 key support
;...e

;...svoid cdecl key_init\40\void\41\:0:
key_init	PROC	USES BX DX DS ES
		LOCAL	shift_status:BYTE

		XOR	AL,AL
		DOSCALL	dos_ctrl_break	; Get current status
		MOV	CS:[old_ctrl_break],DL

		MOV	AL,1
		XOR	DL,DL
		DOSCALL	dos_ctrl_break	; Turn off break checking

		MOV	AL,b_brk
		DOSCALL	dos_get_ivec	; Returned in ES:BX
		MOV	CS:[old_off_b_brk],BX
		MOV	CS:[old_seg_b_brk],ES

		MOV	AX,CS
		MOV	DS,AX
		MOV	DX,OFFSET _key_isr
		MOV	AL,b_brk
		DOSCALL	dos_set_ivec	; Install hacking patch

		MOV	AL,d_brk
		DOSCALL	dos_get_ivec	; Returned in ES:BX
		MOV	CS:[old_off_d_brk],BX
		MOV	CS:[old_seg_d_brk],ES

		MOV	AX,CS
		MOV	DS,AX
		MOV	DX,OFFSET _null_isr
		MOV	AL,d_brk
		DOSCALL	dos_set_ivec	; Install hacking patch
;
; Now decide if BIOS support for 102 key keyboard exists
;
		KEYCALL	key_get_shift	; AL = status
		MOV	shift_status,AL
		XOR	AL,0FFH
		KEYCALL	key_get_shift_x	; AL = extended status (hopefully)
		CMP	AL,shift_status
		JNE	no_extended
		MOV	CS:[extended],1
no_extended:
;
; Remember we are installed
;
		MOV	CS:[inited],1

		RET

key_init	ENDP
;...e
;...svoid cdecl key_deinit\40\void\41\:0:
key_deinit	PROC	USES DX DS

		MOV	AL,CS:[inited]
		AND	AL,AL
		JZ	SHORT not_inited

		MOV	AL,1
		MOV	DL,CS:[old_ctrl_break]
		DOSCALL	dos_ctrl_break	; Restore old break checking flag

		LDS	DX,CS:[old_b_brk]
		MOV	AL,b_brk
		DOSCALL	dos_set_ivec	; Restore old BIOS vector

		LDS	DX,CS:[old_d_brk]
		MOV	AL,d_brk
		DOSCALL	dos_set_ivec	; Restore old DOS vector

not_inited:	RET

key_deinit	ENDP
;...e
;...sint cdecl key_status\40\void\41\:0:
key_status	PROC

		MOV	AL,CS:[extended]
		AND	AL,AL
		JNZ	ksx_loop

;
; Normal support
;

ks_loop:	KEYCALL	key_ready
		JNZ	SHORT ks_ready	; Is a key pending
		XOR	AX,AX		; Return 0, no key ready
		RET

ks_ready:	AND	AX,AX		; Is key Ctrl-Break
		JNZ	SHORT ks_return_it
					; If not then return key
		KEYCALL	key_read_ch	; Swallow key
		JMP	ks_loop

ks_return_it:	AND	AL,AL		; Is it an extended key
		JZ	SHORT ks_func
		XOR	AH,AH		; If normal, ignore scan code
ks_func:	RET

;
; Extended support
;

ksx_loop:	KEYCALL	key_ready_x
		JNZ	SHORT ksx_ready
		XOR	AX,AX
		RET

ksx_ready:	AND	AX,AX		; Is key Ctrl-Break ?
		JNZ	SHORT ksx_return_it
		KEYCALL	key_read_ch_x	; Discard Ctrl-Break
		JMP	ksx_loop	; Get another key

ksx_return_it:	AND	AL,AL		; Is it an normal function key
		JZ	SHORT ksx_func	; If so, return it as 0XX00H
		CMP	AX,000E0H	; Is it Alt-224
		JE	SHORT ksx_norm	; If so, its normal
		CMP	AL,0E0H		; Is is an extended function key
		JE	SHORT ksx_func	; If so, return it as 0XX00H
		XOR	AH,AH		; Otherwise, just return ASCII
ksx_norm:	RET

ksx_func:	XOR	AL,AL
		RET

key_status	ENDP
;...e
;...sint cdecl key_read\40\void\41\:0:
key_read	PROC

		MOV	AL,CS:[extended]
		AND	AL,AL
		JNZ	krx_loop
;
; Normal support
;
kr_loop:	KEYCALL	key_read_ch
		AND	AX,AX		; Is key Ctrl-Break
		JZ	kr_loop		; If so then try next key
		AND	AL,AL		; Is it an extended key
		JZ	SHORT kr_func
		XOR	AH,AH		; If normal, ignore scan code
kr_func:	RET
;
; Extended support
;
krx_loop:	KEYCALL	key_read_ch_x	; Get new key
		AND	AX,AX		; Is it Ctrl-Break
		JZ	krx_loop	; If so, get another
		AND	AL,AL		; Is it a normal function key
		JZ	SHORT krx_func	; If so, return it as 0XX00H
		CMP	AX,000E0H	; Is it Alt-224
		JE	SHORT krx_norm	; if so, normal
		CMP	AL,0E0H		; Is it an extended function key
		JE	SHORT krx_func	; If so, return it as 0XX00H
		XOR	AH,AH		; Otherwise, return the ASCII
krx_norm:	RET

krx_func:	XOR	AL,AL
		RET

key_read	ENDP
;...e
;...sint cdecl key_break\40\void\41\:0:
key_break	PROC

		XOR	AX,AX
		XCHG	AX,CS:[break_flag]
		RET

key_break	ENDP
;...e

;...s_null_isr:0:
;
; New break interrupt server
;
_null_isr	PROC	FAR

		IRET

_null_isr	ENDP
;...e
;...s_key_isr:0:
;
; New break interrupt server
;
_key_isr	PROC	FAR

		MOV	CS:[break_flag],1
		IRET

_key_isr	ENDP
;...e

		END
