/*

th.c - Terminal handler (32 bit DOS with DOS Extender version)

The driver will save the screen if it can. It can be made to refuse to load if
it cannot save the screen by or-ing in MODE_SAVE_REQD to the requested mode.

This code uses assorted int386 and intr calls to perform low level keyboard
input without hanging up on Control-Break.

Enhanced to support a wider range of screen sizes.

	80 columns, 25 lines on MDA machines.
	40 or 80 columns, 25 lines on CGA machines.
	40 or 80 columns, 25 or 43 lines on EGA machines.
	40 or 80 columns, 12,14,21,25,28,43 or 50 lines on VGA machines.
	40,80 or 132 columns, 12,14,21,25,28,43 or 50 lines on XGA2 machines.
	Assorted unusual modes on a variety of super VGA machines
	User extension to override internal tables for debugging.

*/

/*...sincludes:0:*/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <dos.h>
#include <conio.h>
#include <signal.h>
#include <i86.h>
#define	TH_C
#include "th.h"

/*...vcrit\46\h:0:*/
/*...vkey\46\h:0:*/
/*...vth\46\h:0:*/
/*...e*/
/*...stypes:0:*/
typedef unsigned  char  byte;
typedef unsigned short  word;
/*...e*/
/*...sBIOS \35\defines:0:*/
#define	VID		0x10		/* Video BIOS is interrupt 10 hex    */

#define	VID_SET_MODE	0x00		/* Set video mode                    */
#define	VID_SET_CUR_SZ	0x01		/* Set cursor size                   */
#define	VID_SET_CUR_POS	0x02		/* Set cursor position               */
#define	VID_GET_CUR	0x03		/* Get cursor position and size      */
#define	VID_SET_ACT_PGE	0x05		/* Set active display page           */
#define	VID_GET_MODE	0x0f		/* Get current mode                  */
#define	VID_CHAR_GEN	0x11		/* Character generator               */
#define	VID_ALT_SELECT	0x12		/* Alternate select                  */
#define	VID_DISP_COMB	0x1a		/* Display combination code          */

#define	KEY		0x16		/* Keyboard BIOS interrupt           */

#define	KEY_READ_CH	0x00		/* Wait and get character            */
#define	KEY_READY	0x01		/* Test keyboard status              */
#define	KEY_GET_SHIFT	0x02		/* Get shift status                  */
#define	KEY_READ_CH_X	0x10		/* Extended wait and get character   */
#define	KEY_READY_X	0x11		/* Extended test keyboard status     */
#define	KEY_GET_SHIFT_X	0x12		/* Get extended shift status         */
/*...e*/

/*...sget_bios_mode:0:*/
static int get_bios_mode(void)
	{
	union REGS regs;
	regs.h.ah = VID_GET_MODE;
	int386(VID, &regs, &regs);
	return regs.h.al;
	}
/*...e*/
/*...sset_bios_mode:0:*/
static void set_bios_mode(int mode)
	{
	union REGS regs;
	regs.h.ah = VID_SET_MODE;
	regs.h.al = (unsigned char) mode;
	int386(VID, &regs, &regs);
	}
/*...e*/
/*...svmodes:0:*/
/*...sVSUPP type:0:*/
typedef word VSUPP;
#define	VSUPP_MDA	((VSUPP) 0x0001)
#define	VSUPP_CGA	((VSUPP) 0x0002)
#define	VSUPP_EGA	((VSUPP) 0x0004)
#define	VSUPP_VGA	((VSUPP) 0x0008)
#define	VSUPP_XGA2	((VSUPP) 0x0010)
#define	VSUPP_ET4000	((VSUPP) 0x0020)
#define	VSUPP_PARADISE	((VSUPP) 0x0040)
#define	VSUPP_VEGA	((VSUPP) 0x0080)
#define	VSUPP_AHEAD	((VSUPP) 0x0100)
#define	VSUPP_ATI_VGA	((VSUPP) 0x0200)
#define	VSUPP_ATI_EGA	((VSUPP) 0x0400)
#define	VSUPP_TRIDENT	((VSUPP) 0x0800)
#define	VSUPP_OAK	((VSUPP) 0x1000)
#define	VSUPP_USER	((VSUPP) 0x4000)

#define	VSUPP_SVGA	((VSUPP) 0x7fe0)
/*...e*/
/*...svmodes table:0:*/
typedef struct
	{
	int mode;
	int x_size, y_size;
	TH_BOOLEAN colour;
	word segment;
	VSUPP vsupp;	/* Mode is supported for these types of support */
	} VMODE;

/*
Table is arranged so that CGA modes are found before MDA mode 7.
*/

#define	N_VMODES	60

static VMODE vmodes[N_VMODES] =
	{
	0x00,	 40,25,	TH_FALSE,	0xb800,	VSUPP_SVGA|VSUPP_XGA2|VSUPP_VGA|VSUPP_EGA|VSUPP_CGA,
	0x01,	 40,25,	TH_TRUE,	0xb800,	VSUPP_SVGA|VSUPP_XGA2|VSUPP_VGA|VSUPP_EGA|VSUPP_CGA,
	0x02,	 80,25,	TH_FALSE,	0xb800,	VSUPP_SVGA|VSUPP_XGA2|VSUPP_VGA|VSUPP_EGA|VSUPP_CGA,
	0x03,	 80,25,	TH_TRUE,	0xb800,	VSUPP_SVGA|VSUPP_XGA2|VSUPP_VGA|VSUPP_EGA|VSUPP_CGA,

	/* Also listed as colour so if we try to set colour mode something happens */
	0x07,	 80,25,	TH_FALSE,	0xb000,	VSUPP_SVGA|VSUPP_XGA2|VSUPP_VGA|VSUPP_EGA|VSUPP_CGA|VSUPP_MDA,
	0x07,	 80,25,	TH_TRUE,	0xb000,	VSUPP_SVGA|VSUPP_XGA2|VSUPP_VGA|VSUPP_EGA|VSUPP_CGA|VSUPP_MDA,

	0x14,	132,25,	TH_TRUE,	0xb800, VSUPP_XGA2,

	/* Arranged so as to pick modes with base of 0xb800 first */
	0x21,	132,60,	TH_TRUE,	0xb800,	VSUPP_ET4000,
	0x22,	132,44,	TH_TRUE,	0xb800,	VSUPP_ET4000,
	0x23,	132,25,	TH_TRUE,	0xb800,	VSUPP_ET4000,
	0x24,	132,28,	TH_TRUE,	0xb800,	VSUPP_ET4000,
	/* Note the follow is a repeat, but with base of 0xb000 */
	0x18,	132,44,	TH_TRUE,	0xb000,	VSUPP_ET4000,
	0x19,	132,25,	TH_TRUE,	0xb000,	VSUPP_ET4000,
	0x1a,	132,28,	TH_TRUE,	0xb000,	VSUPP_ET4000,
	/* Unusual modes */
	0x26,	 80,60,	TH_TRUE,	0xb800,	VSUPP_ET4000,
	0x2a,	100,40,	TH_TRUE,	0xb800,	VSUPP_ET4000,

	0x21,	132,44,	TH_TRUE,	0xb800,	VSUPP_PARADISE,
	0x51,	 80,30,	TH_TRUE,	0xb800,	VSUPP_PARADISE, /* Mode only exists on EGA-480, assumed 0xb800 */
	0x54,	132,43,	TH_TRUE,	0xb800,	VSUPP_PARADISE, /* Known for VGA, assumed same for EGA-480 */
		/* should the above line be 44 lines high? */
	0x55,	132,25,	TH_TRUE,	0xb800,	VSUPP_PARADISE, /* Known for VGA, assumed same for EGA-480 */

	/* Base address of 0xb800 is assumed, not known in all VEGA modes */
	0x40,	 80,43,	TH_TRUE,	0xb800, VSUPP_VEGA,
	0x41,	132,25,	TH_TRUE,	0xb800, VSUPP_VEGA,
	0x42,	132,43,	TH_TRUE,	0xb800, VSUPP_VEGA,
	0x43,	 80,60,	TH_TRUE,	0xb800, VSUPP_VEGA,
	0x44,	100,60,	TH_TRUE,	0xb800, VSUPP_VEGA,
	0x4d,	120,25,	TH_TRUE,	0xb800, VSUPP_VEGA,
	0x4e,	120,43,	TH_TRUE,	0xb800, VSUPP_VEGA,
	0x4f,	132,25,	TH_TRUE,	0xb800, VSUPP_VEGA,

	/* Base addresses may be wrong for EGA2001 */
	0x22,	132,44,	TH_TRUE,	0xb800, VSUPP_AHEAD,
	0x23,	132,25,	TH_TRUE,	0xb800, VSUPP_AHEAD,
	/* Mode does not exist for EGA2001 - oops! */
	0x24,	132,28,	TH_TRUE,	0xb800, VSUPP_AHEAD,

	0x08,	132,25,	TH_TRUE,	0xb800, VSUPP_ATI_VGA,

	0x23,	132,25,	TH_TRUE,	0xb800, VSUPP_ATI_EGA,
	0x33,	132,44,	TH_TRUE,	0xb800, VSUPP_ATI_EGA,
	0x58,	80,33,	TH_TRUE,	0xb800, VSUPP_ATI_EGA,

	/* Note the duplication of the 80x43 EGA/VGA mode */
	0x50,	80,30,	TH_TRUE,	0xb800, VSUPP_TRIDENT,
	0x51,	80,43,	TH_TRUE,	0xb800, VSUPP_TRIDENT,
	0x52,	80,60,	TH_TRUE,	0xb800, VSUPP_TRIDENT,
	/* This block have 9 pixel wide characters */
	0x57,	132,25,	TH_TRUE,	0xb800, VSUPP_TRIDENT,
	0x58,	132,30,	TH_TRUE,	0xb800, VSUPP_TRIDENT,
	0x59,	132,43,	TH_TRUE,	0xb800, VSUPP_TRIDENT,
	0x5a,	132,60,	TH_TRUE,	0xb800, VSUPP_TRIDENT,
	/* This block have 8 pixel wide characters */
	0x53,	132,25,	TH_TRUE,	0xb800, VSUPP_TRIDENT,
	0x54,	132,30,	TH_TRUE,	0xb800, VSUPP_TRIDENT,
	0x55,	132,43,	TH_TRUE,	0xb800, VSUPP_TRIDENT,
	0x56,	132,60,	TH_TRUE,	0xb800, VSUPP_TRIDENT,

	0x50,	132,25,	TH_TRUE,	0xb800, VSUPP_OAK,
	0x51,	132,43,	TH_TRUE,	0xb800, VSUPP_OAK,

/* 48 modes upto here, space for 12 more */

	};

static int n_vmodes = 48;
/*...e*/
/*...sappend_vmode:0:*/
static void append_vmode(
	int mode, int x_size, int y_size, TH_BOOLEAN colour,
	word segment, VSUPP vsupp
	)
	{
	if ( n_vmodes == N_VMODES )
		return; /* Silently ignore */
	vmodes[n_vmodes].mode    = mode   ;
	vmodes[n_vmodes].x_size  = x_size ;
	vmodes[n_vmodes].y_size  = y_size ;
	vmodes[n_vmodes].colour  = colour ;
	vmodes[n_vmodes].segment = segment;
	vmodes[n_vmodes].vsupp   = vsupp  ;
	n_vmodes++;
	}
/*...e*/
/*...sget_vsupp:0:*/
/*
Determine what we think we are capable of, by making test calls to BIOS etc.
However if there is no EGA, VGA or better, and we are in mode 7,
then assume this is because there are no CGA modes 0-3 to be in instead.
*/

/*...sget_vsupp_bios:0:*/
/*...svsupp_user:0:*/
static TH_BOOLEAN vsupp_user(void)
	{
	char	*s;
	int	mode, x_size, y_size;
	word	segment;

	if ( (s = getenv("TH_VSUPP_USER")) == NULL )
		return TH_FALSE;

	mode = (int) strtol(s, &s, 0);
	if ( s == NULL || *s++ != ',' )
		return TH_FALSE;
	x_size = (int) strtol(s, &s, 0);
	if ( s == NULL || *s++ != ',' )
		return TH_FALSE;
	y_size = (int) strtol(s, &s, 0);
	if ( s == NULL || *s++ != ',' )
		return TH_FALSE;
	segment = (int) strtol(s, &s, 0);
	if ( s == NULL )
		return TH_FALSE;

	/* Append a new mode to our internal table */

	append_vmode(mode, x_size, y_size, TH_TRUE, segment, VSUPP_USER);

	return TH_TRUE;
	}
/*...e*/
/*...scheck_for_reg:0:*/
/*...swrite_reg:0:*/
static void write_reg(word port, byte reg_num, byte value)
	{
	outpw(port,((word)value<<8)|reg_num);
	}
/*...e*/
/*...sread_reg:0:*/
static byte read_reg(word port, byte reg_num)
	{	
	outp(port,reg_num);
	return (byte) inp(port+1);
	}
/*...e*/

static TH_BOOLEAN check_for_reg(word port, byte reg_num, byte bits)
	{
	byte old_reg = read_reg(port,reg_num);

	write_reg(port,reg_num,bits);
	if ( (read_reg(port,reg_num) & bits) == bits )
		/* Can set all of register bits */
		{
		write_reg(port,reg_num,0);
		if ( (read_reg(port,reg_num) & bits) == 0 )
			/* Can clear all of register bits */
			{
			write_reg(port,reg_num,old_reg);
			return TH_TRUE;
			}
		}
		
	write_reg(port,reg_num,old_reg);
	return TH_FALSE;
	}
/*...e*/

static VSUPP get_vsupp_bios(void)
	{
	union REGS regs;

/*...suser:8:*/
if ( vsupp_user() )
	return VSUPP_USER;
/*...e*/
/*...sCGA only:8:*/
regs.h.ah = VID_ALT_SELECT;
regs.h.bl = 0x10; /* Return EGA information */
int386(VID, &regs, &regs);
if ( (regs.h.bl & 0xfc) || regs.h.cl != 0x09 )
	return VSUPP_CGA;
/*...e*/
/*...sEGA only:8:*/
regs.h.ah = VID_DISP_COMB;
regs.h.al = 0x00; /* VGA: Read display combination code */
int386(VID, &regs, &regs);
if ( regs.h.al != 0x1a )
	return VSUPP_EGA;
/*...e*/
#if 0
/*...sXGA2:8:*/
/* Look for an XGA BIOS support for mode 0x14, 132 column modes */
{
byte buf[0x40];
word *s  = (word *) buf; /* State buffer */
byte *s2 = (byte *) buf; /* State buffer */
struct SREGS sregs;

regs.x.eax = 0x1b00; /* Return functionality / state information */
regs.x.ebx = 0x0000; /* Implementation type 0 */
regs.x.edi = FP_OFF(s);
sregs.es   = FP_SEG(s);
int386x(VID, &regs, &regs, &sregs);
if ( regs.h.al == 0x1b )
	/* BIOS function supported */
	{
	byte *sft; /* Static functionality table */

	sft = (byte *) ( (((unsigned) s[1]) << 4) + (unsigned) s[0] );

	if ( (sft[2] & 0x10) != 0 )
		/* Mode 0x14 supported */
		return VSUPP_XGA2;

	/* In early internal drops of the XGA2 BIOS, the above test fails,
	   despite being the correct architected test. The following test
	   should also be true for XGA2 systems and means the same thing. */

	if ( (s2[0x2e] & 0x10) != 0 )
		/* 132 column mode supported */
		return VSUPP_XGA2;
	}
}
/*...e*/
#endif
/*...sTseng:8:*/
regs.w.ax = 0x10f1; /* Tseng ET4000 chipset, Get DAC type */
int386(VID, &regs, &regs);
if ( regs.h.al == 0x10 )
	return VSUPP_ET4000;
/*...e*/
/*...sParadise:8:*/
/* The following test will not differeniate between "Paradise EGA-480"
   and "Paradise VGA" cards. However, the EGA-480 card (for which I
   have no information) supports a different set of modes to the VGA */

{
char *sig = (char *) 0x000c007d;
if ( !memcmp(sig, "VGA=", 4) )
	return VSUPP_PARADISE;
}
/*...e*/
/*...sVEGA:8:*/
regs.w.ax = 0x6f00; /* VEGA / Video 7 extensions installation check */
int386(VID, &regs, &regs);
if ( regs.w.bx == 0x5637 )
	return VSUPP_VEGA;
/*...e*/
/*...sAhead:8:*/
/* The following test will not differentiate between the "Ahead
   Systems EGA2001", "Ahead B" and "Ahead B (Wizard/3270)" cards.
   However, the EGA2001 (for which I have insufficient information),
   looks as if it is a subset of the Ahead B modes, and has the same
   mode numbers, and probably the same base addresses. The Ahead
   Wizard 3270 has different mode numbers and unusual screen sizes.
   I will not support it here. */

{
char *sig = (char *) 0x000c0025;
if ( !memcmp(sig, "AHEAD", 5) )
	return VSUPP_AHEAD;
}
/*...e*/
/*...sATI:8:*/
/* The following test will detect the "ATI EGA Wonder" (assumed 800+),
   and will drive it in its 132x25 and 132x44 modes. This test may
   accidentally match an "ATI VIP". If this happens, all will be ok,
   as this has the same modes. */

{
char *sig = (char *) 0x000c0031;
if ( !memcmp(sig, "761295520", 9) )
	{
	sig = (char *) 0x000c0040;
	if ( !memcmp(sig, "31", 2) )
		{
		regs.w.ax = 0x5555;
		int386(VID, &regs, &regs);
		if ( regs.w.ax == 0xaaaa )
			return VSUPP_ATI_VGA;
		}
	if ( !memcmp(sig, "32", 2) )
		return VSUPP_ATI_EGA;
	}
}
/*...e*/
/*...sTrident:8:*/
{
int old_reg, new_reg;
outp(0x03c4, 0x0e);		/* Mode Control Register */
old_reg = inp(0x03c5);		/* Save current value */
outp(0x03c5, 0x00);		/* Write all zeroes */
new_reg = inp(0x03c5);		/* Get new value */
outp(0x03c5, 0x02);		/* Set bank 0 (use save_reg here?) */
if ( (new_reg & 0x0f) == 0x02 )
	return VSUPP_TRIDENT;
}
/*...e*/
/*...sOak:8:*/
if ( check_for_reg(0x03de, 0x11, 0xff) )
	/* See if bank switch register exists */
	return VSUPP_OAK;
/*...e*/

	return VSUPP_VGA;
	}
/*...e*/

static VSUPP get_vsupp(void)
	{
	VSUPP vsupp = get_vsupp_bios();

	if ( get_bios_mode() == 7 && vsupp == VSUPP_CGA )
		vsupp = VSUPP_MDA;

	return vsupp;
	}
/*...e*/

static VSUPP vsupp;

/*...sindex_of_mode:0:*/
static int index_of_mode(int mode)
	{
	int i;
	for ( i = 0; i < n_vmodes; i++ )
		if ( (vmodes[i].vsupp & vsupp) != 0 &&
		     vmodes[i].mode == mode )
			return i;
	return -1;
	}
/*...e*/
/*...slookup_mode:0:*/
static int lookup_mode(int x_size, int y_size, TH_BOOLEAN colour)
	{
	int i;
	for ( i = 0; i < n_vmodes; i++ )
		if ( (vmodes[i].vsupp & vsupp) != 0 &&
		     vmodes[i].x_size == x_size &&
		     vmodes[i].y_size == y_size &&
		     vmodes[i].colour == colour )
			return i;
	return -1;
	}
/*...e*/
/*...e*/
/*...sget_mode:0:*/
static TH_BOOLEAN get_mode(int *x_size, int *y_size, TH_BOOLEAN *colour)
	{
	int i, mode = get_bios_mode();

	if ( (i = index_of_mode(mode)) == -1 )
		/* Mode not recognised as a mode for level of support */
		return TH_FALSE;

	*x_size = vmodes[i].x_size;
	*colour = vmodes[i].colour;

	if ( vsupp & (VSUPP_EGA|VSUPP_VGA|VSUPP_XGA2|VSUPP_SVGA) )
		/* Can be many sizes for a given mode, so query current */
		{
		union REGS regs;

		regs.h.ah = VID_CHAR_GEN;
		regs.h.al = 0x30; /* Font information */
		regs.h.bh = 0x00; /* Get current INT 1F pointer */
		int386(VID, &regs, &regs);
		*y_size = regs.h.dl + 1;
		}
	else
		*y_size = vmodes[i].y_size;

	return TH_TRUE;
	}
/*...e*/
/*...sset_mode:0:*/
#define	SCAN_200	0x00
#define	SCAN_350	0x01
#define	SCAN_400	0x02

#define	FONT_8		0x12
#define	FONT_14		0x11
#define	FONT_16		0x14

/*...sset_bios_mode_ega:0:*/
/*
Scan lines on the screen is 350.
Fonts are 14 or 8 scan lines high.
Implies 350/14 (=25) or 350/8 (=43) lines allowable.
The BIOS VID_ALT_SELECT call is not available in EGA systems, therefore a
200 scan line high mode cannot be acheived. Microsoft directly reprogram the
CRTC in their OS/2 Family API support library API.LIB, allowing them to get
a 200/14 (=14) line mode. We will refrain from such black magic.
*/

static void set_bios_mode_ega(int mode, int y_size)
	{
	union REGS regs;
	unsigned char font;

	switch ( y_size )
		{
		case 25: font = FONT_14; break;
		case 43: font = FONT_8;  break;
		}

	set_bios_mode(mode);

	regs.h.ah = VID_CHAR_GEN;
	regs.h.al = font;
	regs.h.bl = 0;
	int386(VID, &regs, &regs);

/*...sEGA BIOS cursor bug:8:*/
regs.h.ah = VID_SET_CUR_SZ;
regs.h.ch = 0x06;
regs.h.cl = 0x07;
int386(VID, &regs, &regs);
/*...e*/
	}
/*...e*/
/*...sset_bios_mode_vga:0:*/
/*
Scan lines on the screen are 200, 350 or 400.
Fonts are 14, 8 and 16 scan lines high.
Implies the following numbers of lines selectable :-

	    |	200	350	400
	----+----------------------
	8   |	25	43	50
	14  |	14	25	28
	16  |	12	21	25

Although this function is marked as VGA, it can also be used for the XGA2
mode 0x14.
*/

static TH_BOOLEAN set_bios_mode_vga(int mode, int y_size)
	{
	union REGS regs;
	unsigned char scans, font;

	switch ( y_size )
		{
		case 12: scans = SCAN_200; font = FONT_16; break;
		case 14: scans = SCAN_200; font = FONT_14; break;
		case 21: scans = SCAN_350; font = FONT_16; break;
		case 25: scans = SCAN_400; font = FONT_16; break;
		case 28: scans = SCAN_400; font = FONT_14; break;
		case 43: scans = SCAN_350; font = FONT_8;  break;
		case 50: scans = SCAN_400; font = FONT_8;  break;
		}

	regs.h.ah = VID_SET_ACT_PGE;
	regs.h.al = 0; /* Active page 0 must be in force for next call */
	int386(VID, &regs, &regs);

	regs.h.ah = VID_ALT_SELECT;
	regs.h.bl = 0x30; /* Select scan lines */
	regs.h.al = scans;
	int386(VID, &regs, &regs);
	if ( regs.h.al != 0x12 )
		return TH_FALSE;

	set_bios_mode(mode);

/*...sidle for a little while to fix BIOS bug:8:*/
{
int i, j = 0;
for ( i = 0; i < 10000; i++ )
	j = (j * j + j) ^ j;
}
/*...e*/

	regs.h.ah = VID_CHAR_GEN;
	regs.h.al = font;
	regs.h.bl = 0;
	int386(VID, &regs, &regs);

	return TH_TRUE;
	}
/*...e*/

static TH_BOOLEAN set_mode(int x_size, int y_size, TH_BOOLEAN colour)
	{
	int y_size_look_for = y_size;
	int i;

	if ( vsupp == VSUPP_EGA &&
	     (x_size == 40 || x_size == 80) &&
	     (y_size == 43) )
		y_size_look_for = 25;

	if ( (vsupp & (VSUPP_VGA|VSUPP_XGA2|VSUPP_SVGA)) != 0 &&
	     (x_size == 40 || x_size == 80) &&
	     (y_size == 12 || y_size == 14 || y_size == 21 ||
	      y_size == 28 || y_size == 43 || y_size == 50) )
		y_size_look_for = 25;

	if ( vsupp == VSUPP_XGA2 &&
	     (x_size == 132) &&
	     (y_size == 12 || y_size == 14 || y_size == 21 ||
	      y_size == 28 || y_size == 43 || y_size == 50) )
		y_size_look_for = 25;

	if ( (i = lookup_mode(x_size, y_size_look_for, colour)) == -1 )
		/* Can't find a mode to support the desired size, */
		/* given the level of support indentified earlier */
		return TH_FALSE;

	switch ( vsupp )
		{
/*...sVSUPP_MDA             \45\ only mode 7 exists:16:*/
/*
Can only have a 25 line mode screen.
This must be mode 7, and we are already in it.
*/

case VSUPP_MDA:
	break;
/*...e*/
/*...sVSUPP_CGA             \45\ no EGA or VGA support:16:*/
case VSUPP_CGA:
	set_bios_mode(vmodes[i].mode);
	break;
/*...e*/
/*...sVSUPP_EGA             \45\ EGA\44\ but not VGA BIOS support found:16:*/
/*
Uses character generator functions to get at unusual modes.
*/

case VSUPP_EGA:
	set_bios_mode_ega(vmodes[i].mode, y_size);
	break;
/*...e*/
/*...sVSUPP_VGA\44\ VSUPP_XGA2 \45\ VGA BIOS or XGA2 BIOS:16:*/
/*
Uses character generator functions, and also sets scan lines on display to get
at unusual modes.
*/

case VSUPP_VGA:
case VSUPP_XGA2:
	if ( !set_bios_mode_vga(vmodes[i].mode, y_size) )
		return TH_FALSE;
	break;
/*...e*/
/*...sdefault               \45\ something better \40\eg\58\ SVGA\41\:16:*/
/*
If the mode is an additional SVGA type mode, then simply set it.
Else attempt to use VGA level support to set mode with its trickery.
*/

default:
	{
	int mode = vmodes[i].mode;

	if ( mode > 3 )
		set_bios_mode(mode);
	else
		{
		if ( !set_bios_mode_vga(mode, y_size) )
			return TH_FALSE;
		}
	}
	break;
/*...e*/
		}
	return TH_TRUE;
	}		
/*...e*/
/*...sptr_to_screen:0:*/
static word *ptr_to_screen(void)
	{
	word *sc_ptr;
	int mode = get_bios_mode();
	int i = index_of_mode(mode);
	sc_ptr = (word *) ( (unsigned) vmodes[i].segment << 4 );
	return sc_ptr;
	}
/*...e*/

/*...svars:0:*/
int x_size;				/* Width of screen                   */
int y_size;				/* Height of screen                  */

static int tmp;				/* Holds (y_size - 1) * x_size       */
static word *sc_ptr;			/* Pointer that points into screen   */
static int csr_type = CSR_OFF;		/* what sort of cursor do we have    */
static int csr_x;			/* current x in range 0 to y_size-1  */
static int csr_y;			/* current y in range 0 to x_size-1  */
static word csr_col_00;			/* current colour * 0x100 for speed  */
#define	BLANK	(csr_col_00 + ' ')	/* word used for blanking screen etc */
/*...e*/
/*...sscmove:0:*/
static void scmove(int src, int dest, unsigned int count)
	{
	memcpy(sc_ptr + dest, sc_ptr + src, count * sizeof(word));
	}
/*...e*/
/*...sscfill:0:*/
static void scfill(int dest, word val, register unsigned int count)
	{
	word *p = sc_ptr + dest;
	while ( count-- )
		*p++ = val;
	}
/*...e*/

/*...sset_cursor_size:0:*/
static void set_cursor_size(int top, int bottom)
	{
	union REGS regs;
	regs.h.ah = VID_SET_CUR_SZ;
	regs.h.ch = (unsigned char) top;
	regs.h.cl = (unsigned char) bottom;
	int386(VID, &regs, &regs);
	}
/*...e*/
/*...sset_cursor_position:0:*/
static union REGS inregs_scp;

static void set_cursor_position(register int x, register int y)
	{
	if ( csr_type != CSR_OFF )
		{
		union REGS regs;
		inregs_scp.h.dl = (unsigned char) x;
		inregs_scp.h.dh = (unsigned char) y;
		int386(VID, &inregs_scp, &regs);
		}
	}
/*...e*/

/*...sth_beep:0:*/
void th_beep(void)
	{
	union REGS regs;
	regs.w.ax = 0x0e07;
	int386(VID, &regs, &regs);
	}
/*...e*/
/*...sth_cxy:0:*/
void th_cxy(int x, int y)
	{
	set_cursor_position(csr_x = x, csr_y = y);
	}
/*...e*/
/*...sth_cy:0:*/
void th_cy(int y)
	{
	set_cursor_position(csr_x = 0, csr_y = y);
	}
/*...e*/
/*...sth_chome:0:*/
void th_chome(void)
	{
	set_cursor_position(csr_x = 0, csr_y = 0);
	}
/*...e*/
/*...sth_cls:0:*/
void th_cls(void)
	{
	scfill(0, BLANK, y_size * x_size);
	set_cursor_position(csr_x = 0, csr_y = 0);
	}
/*...e*/
/*...sth_setcol:0:*/
void th_setcol(int col)
	{
	csr_col_00 = (col << 8);
	}
/*...e*/
/*...sth_setcsr:0:*/
/*
The following set of values have been carefully chosen
so as to look good on both CGA and EGA cards, AND also
on EGA cards simulating CGA text modes. Alter at your peril!
*/

#define	CS_MONO	0
#define	CS_EMUL	1
#define	CS_EGA	2
static int csr_inx;

static int pos_alt   [] = {  6, 4, 2 };
static int pos_uline [] = { 12, 6, 4 };
static int pos_bottom[] = { 12, 7, 7 };

void th_setcsr(int csr)
	{
	if ( csr == csr_type )
		return;

	if ( csr == CSR_OFF )
		{
		set_cursor_position(0, 60);
		csr_type = csr;
		}
	else
		{
		switch ( csr )
			{
			case CSR_BLOCK:
				set_cursor_size(0, pos_bottom[csr_inx]);
				break;
			case CSR_ULINE:
				set_cursor_size(pos_uline[csr_inx], pos_bottom[csr_inx]);
				break;
			case CSR_ALT:
				set_cursor_size(pos_alt[csr_inx], pos_bottom[csr_inx]);
				break;
			}
	
		if ( csr_type == CSR_OFF )
			{
			csr_type = csr;
			set_cursor_position(csr_x, csr_y);
			}
		else
			csr_type = csr;
		}
	}
/*...e*/
/*...sth_inslin:0:*/
void th_inslin(void)
	{
	int i;
	register int p, q;

	p = tmp;
	for ( i = y_size - 1; i > csr_y; i-- )
		{
		scmove(q = (p - x_size), p, x_size); p = q;
		}
	scfill(csr_y * x_size, BLANK, x_size);
	}
/*...e*/
/*...sth_dellin:0:*/
void th_dellin(void)
	{
	register int p;
	p = csr_y * x_size;
	scmove(p + x_size, p, tmp - p);
	scfill(tmp, BLANK, x_size);
	}
/*...e*/
/*...sth_deleol:0:*/
void th_deleol(void)
	{
	register int p;
	p = csr_y * x_size + csr_x;
	scfill(p, BLANK, x_size - csr_x);
	}
/*...e*/
/*...sth_save:0:*/
void th_save(int x, int y, int n, void *buf)
	{
	word *p = sc_ptr + (y * x_size + x);
	memcpy(buf, p, n*sizeof(word));
	}
/*...e*/
/*...sth_restore:0:*/
void th_restore(int x, int y, int n, const void *buf)
	{
	word *p = sc_ptr + (y * x_size + x);
	memcpy(p, buf, n*sizeof(word));
	}
/*...e*/
/*...sth_isprint:0:*/
/* All characters except NUL are printable. */

TH_BOOLEAN th_isprint(int c)
	{
	return c != 0;
	}
/*...e*/
/*...sth_p_raw_chr:0:*/
void th_p_raw_chr(register int c)
	{
	word *p = sc_ptr + (csr_y * x_size + csr_x);
	*p = csr_col_00 + ((unsigned char) c);
	if ( ++csr_x == x_size )
		{
		csr_x = 0; csr_y++;
		}
	set_cursor_position(csr_x, csr_y);
	}
/*...e*/
/*...sth_p_raw_str:0:*/
void th_p_raw_str(register const char *s)
	{
	word *p = sc_ptr + (csr_y * x_size + csr_x);
	while ( *s )
		{
		*p++ = csr_col_00 + *((const unsigned char *) s++);
		if ( ++csr_x == x_size )
			{
			csr_x = 0; csr_y++;
			p = sc_ptr + (csr_y * x_size);
			}
		}
	set_cursor_position(csr_x, csr_y);
	}
/*...e*/
/*...sth_p_chr:0:*/
void th_p_chr(register int c)
	{
	if ( c == '\n' )
		set_cursor_position(csr_x = 0, ++csr_y);
	else if ( c == '\t' )
		{
		if ( (csr_x = ((csr_x+8) & 0xf8)) == x_size )
			{
			csr_x = 0; csr_y++;
			}
		set_cursor_position(csr_x, csr_y);
		}
	else if ( c == '\b' )
		{
		if ( csr_x )
			csr_x--;
		else
			{
			csr_x = x_size - 1;
			csr_y--;
			}
		set_cursor_position(csr_x, csr_y);
		}
	else if ( c == '\r' )
		set_cursor_position(csr_x = 0, csr_y);
	else
		th_p_raw_chr(c);
	}
/*...e*/
/*...sth_p_str:0:*/
void th_p_str(register const char *s)
	{
	while ( *s )
		th_p_chr(*s++);
	}
/*...e*/
/*...sth_sync_screen:0:*/
void th_sync_screen(void)
	{
	}
/*...e*/
/*...skeyboard stuff:0:*/
static TH_BOOLEAN break_occurred = TH_FALSE;
static TH_BOOLEAN extended;

/*...sth_key_init:0:*/
/* Decide if BIOS support for 102 key keyboards exists */
static void th_key_init(void)
	{
	union REGS regs;
	unsigned char shift_state;
	regs.h.ah = KEY_GET_SHIFT;
	int386(KEY, &regs, &regs);
	shift_state = regs.h.al;
	regs.h.ah = KEY_GET_SHIFT;
	regs.h.al ^= 0xff;
	int386(KEY, &regs, &regs);
	extended = ( regs.h.al == shift_state );
	}
/*...e*/
/*...sth_key_status:0:*/
int th_key_status(void)
	{
	union REGPACK regs;
	regs.x.ds = 0;
	regs.x.es = 0;
	regs.x.fs = 0;
	regs.x.gs = 0;
	if ( extended )
		{
		for ( ;; )
			{
			regs.h.ah = KEY_READY_X;
			intr(KEY, &regs);
			if ( regs.x.flags & INTR_ZF )
				return 0;
			if ( regs.w.ax != 0 )
				break;
			regs.h.ah = KEY_READ_CH_X;
			intr(KEY, &regs);
			}
		if ( regs.h.al == 0 )
			return regs.w.ax & 0xff00;
		if ( regs.w.ax == 0x00e0 )
			return regs.w.ax;
		if ( regs.h.al == 0xe0 )
			return regs.w.ax & 0xff00;
		return regs.h.al;
		}
	else
		{
		for ( ;; )
			{
			regs.h.ah = KEY_READY;
			intr(KEY, &regs);
			if ( regs.x.flags & INTR_ZF )
				return 0;
			if ( regs.w.ax != 0 )
				break;
			regs.h.ah = KEY_READ_CH;
			intr(KEY, &regs);
			}
		if ( regs.h.al == 0 )
			return regs.w.ax;
		return regs.h.al;
		}
	}
/*...e*/
/*...sth_key_read:0:*/
int th_key_read(void)
	{
	union REGS regs;
	if ( extended )
		{
		do
			{
			regs.h.ah = KEY_READ_CH_X;
			int386(KEY, &regs, &regs);
			}
		while ( regs.w.ax == 0 );
		if ( regs.h.al == 0 )
			return regs.w.ax & 0xff00;
		if ( regs.w.ax == 0x00e0 )
			return regs.w.ax;
		if ( regs.h.al == 0xe0 )
			return regs.w.ax & 0xff00;
		return regs.h.al;
		}
	else
		{
		do
			{
			regs.h.ah = KEY_READ_CH;
			int386(KEY, &regs, &regs);
			}
		while ( regs.w.ax == 0 );
		if ( regs.h.al == 0 )
			return regs.w.ax & 0xff00;
		return regs.h.al;
		}
	}
/*...e*/

/*...sth_key_break:0:*/
TH_BOOLEAN th_key_break(void)
	{
	if ( !break_occurred )
		return TH_FALSE;
	break_occurred = TH_FALSE;
	return TH_TRUE;
	}
/*...e*/
/*...e*/
/*...sth_init and th_deinit:0:*/
/*...scatch_sigbreak:0:*/
static void catch_sigbreak()
	{
	break_occurred = TH_TRUE;
	signal(SIGBREAK, catch_sigbreak);
	}
/*...e*/
/*...scatch_sigint:0:*/
static void catch_sigint()
	{
	break_occurred = TH_TRUE;
	signal(SIGINT, catch_sigint);
	}
/*...e*/

static TH_BOOLEAN th_inited = TH_FALSE;

/*...ssaved state:0:*/
static int saved_x_size;		/* Saved columns on startup          */
static int saved_y_size;		/* Saved rows on startup             */
static TH_BOOLEAN saved_colour;		/* Saved 'in-colour' flag on startup */
static byte *saved_buf;			/* Contents of screen on startup     */
static int saved_csr_x;			/* Saved cursor position on startup  */
static int saved_csr_y;			/* Saved cursor position on startup  */
static int saved_csr_top;		/* Saved top line of cursor          */
static int saved_csr_bottom;		/* Saved bottom line of cursor       */
/*...e*/

TH_BOOLEAN th_init(int th_mode, int desired_x_size, int desired_y_size)
	{
	vsupp = get_vsupp();

	if ( !get_mode(&saved_x_size, &saved_y_size, &saved_colour) )
		/* Failed 'cos were not in a known text mode */
		{
		if ( th_mode & MODE_SAVE_REQD )
			return TH_FALSE;
		if ( !set_mode(saved_x_size = 80, saved_y_size = 25, saved_colour = TH_TRUE) )
			return TH_FALSE;
		}

/*...ssave current screen:8:*/
{
union REGS regs;

if ( (saved_buf = (byte *) malloc(saved_x_size * saved_y_size * sizeof(word))) == NULL )
	return TH_FALSE;
memcpy(saved_buf, ptr_to_screen(), saved_x_size * saved_y_size * sizeof(word));

regs.h.ah = VID_GET_CUR;
regs.h.bh = 0; /* Active page assumed 0 */
int386(VID, &regs, &regs);
saved_csr_x      = regs.h.dl;
saved_csr_y      = regs.h.dh;
saved_csr_top    = regs.h.ch;
saved_csr_bottom = regs.h.cl;
}
/*...e*/

	x_size = ( desired_x_size != 0 ) ? desired_x_size : saved_x_size;
	y_size = ( desired_y_size != 0 ) ? desired_y_size : saved_y_size;

	if ( x_size != saved_x_size || y_size != saved_y_size || !saved_colour )
		if ( !set_mode(x_size, y_size, TH_TRUE) )
			return TH_FALSE;

/*...sthings to accelerate execution later:8:*/
{
sc_ptr = ptr_to_screen();
tmp = (y_size - 1) * x_size;
inregs_scp.h.ah = VID_SET_CUR_POS;
inregs_scp.h.bh = 0; /* Active page to use is 0 */
}
/*...e*/

	th_key_init();

	if ( signal(SIGINT, SIG_IGN) != SIG_IGN )
		signal(SIGINT, catch_sigint);
	if ( signal(SIGBREAK, SIG_IGN) != SIG_IGN )
		signal(SIGBREAK, catch_sigbreak);

/*...sdecide on cursor sizes:8:*/
/*
Cursor sizes :-
	MDA	0-13				Monochrome cursor size
	CGA	0-7				No emulation required
	EGA	0-7 (43 line) 0-13 (25 line)	Cursor emulation (buggy)
	VGA	0-7				Scaled by cursor emulation
*/

{
if ( get_bios_mode() == 7 )
	csr_inx = CS_MONO;
else if ( vsupp == VSUPP_EGA )
	/* Tricky case, cursor emulation is unreliable on EGA */
	{
	if ( y_size == saved_y_size )
		csr_inx = CS_EMUL; /* Assume emulation correctly set up */
	else if ( y_size == 25 )
		csr_inx = CS_MONO; /* Emulation will not have been enabled */
	else
		csr_inx = CS_EGA;  /* Special set (emulation in effect when shouldn't be) */
	}
else
	csr_inx = CS_EMUL;
}
/*...e*/

	th_setcsr(CSR_BLOCK);
	th_setcol(COL(COL_WHITE, COL_BLACK, TH_FALSE, TH_FALSE));
	th_cls();

	return th_inited = TH_TRUE;
	}

void th_deinit(void)
	{
	if ( !th_inited )
		return;

/*...srestore current screen:8:*/
{
if ( x_size != saved_x_size || y_size != saved_y_size )
	set_mode(saved_x_size, saved_y_size, saved_colour);

memcpy(ptr_to_screen(), saved_buf, saved_x_size * saved_y_size * sizeof(word));
free((char *) saved_buf);

csr_type = CSR_BLOCK;
set_cursor_position(saved_csr_x, saved_csr_y);
set_cursor_size(saved_csr_top, saved_csr_bottom);
}
/*...e*/

	signal(SIGINT, SIG_DFL);
	signal(SIGBREAK, SIG_DFL);

	th_inited = TH_FALSE;
	}
/*...e*/
