/*

th.c - Terminal handler (EFI)

No beep.
Only a sad subset of keycodes supported.
No support for control break.
Bottom right character must be avoided.
No way to change cursor size.

*/

/*...sincludes:0:*/
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <efi.h>
#include <efilib.h>
#include <efilibc.h>
#define	TH_C
#include "th.h"

/*...vth\46\h:0:*/
/*...e*/

/*...svars:0:*/
int x_size, y_size;

static int init_x_size, init_y_size;
static int csr_x;
static int csr_y;
static unsigned char csr_col = COL(COL_WHITE,COL_BLACK,TH_FALSE);
static int csr_type = CSR_BLOCK;
static unsigned char *screen;

#define	L_BUF 200
static CHAR16 char_buf[1+1];
static CHAR16 line_buf[L_BUF+1];
static CHAR16 wipe_buf[L_BUF+1];

#define In  ST->ConIn
#define	Out ST->ConOut
/*...e*/

/*...swipe:0:*/
static void wipe(int x, int y, int w)
	{
	unsigned char *p = screen + (y*x_size+x)*2;
	int i;
	for ( i = 0; i < w; i++ )
		{
		*p++ = (unsigned char) ' ';
		*p++ = csr_col;
		}

	/* Avoid writing to bottom right */
	if ( y == y_size-1 && x+w == x_size && w > 0 )
		--w;

	Out->SetCursorPosition(Out, x, y);
	Out->SetAttribute(Out, csr_col);
	Out->OutputString(Out, wipe_buf+L_BUF-w);
	}
/*...e*/
/*...srestore:0:*/
/* Make a part of a line of the screen match the internal memory version.
   Output spans of the same attribute in one go, to try to speed up things.
   Unfortunately, despite this optimisation, things still run quite slow.
   Scrolling (inserting and deleting lines) is particularly affected. */

static void restore(int x, int y, int w)
	{
	unsigned char *p = screen + (y*x_size+x)*2;
	Out->SetCursorPosition(Out, x, y);

	if ( y < y_size-1 )
		{
		while ( w-- )
			{
			int n = 1;
			unsigned char col;
			line_buf[0] = *p++;
			col = *p++;
			while ( w > 0 && p[1] == col )
				{
				line_buf[n++] = p[0];
				p += 2;
				--w;
				}
			line_buf[n] = (CHAR16) 0;
			Out->SetAttribute(Out, col);
			Out->OutputString(Out, line_buf);
			}
		}
	else
		/* Slower version, having to take into account that we
		   must avoid overwriting the bottom right character. */
		{
		while ( w-- )
			{
			int n = 1;
			unsigned char col;
			line_buf[0] = *p++;
			col = *p++;
			while ( w > 0 && p[1] == col )
				{
				line_buf[n++] = p[0];
				p += 2;
				--w;
				}
			line_buf[n] = (CHAR16) 0;
			Out->SetAttribute(Out, col);

			/* Avoid writing to bottom right */
			if ( x+n == x_size )
				line_buf[n-1] = (CHAR16) 0;

			Out->OutputString(Out, line_buf);
			x += n;
			}
		}
	}
/*...e*/

/*...sth_beep:0:*/
void th_beep(void)
	{
	}
/*...e*/
/*...sth_cxy:0:*/
void th_cxy(int x, int y)
	{
	csr_x = x;
	csr_y = y;
	}
/*...e*/
/*...sth_cy:0:*/
void th_cy(int y)
	{
	csr_x = 0;
	csr_y = y;
	}
/*...e*/
/*...sth_chome:0:*/
void th_chome(void)
	{
	csr_x = 0;
	csr_y = 0;
	}
/*...e*/
/*...sth_cls:0:*/
void th_cls(void)
	{
	unsigned char *p = screen;
	int i;
	for ( i = 0; i < x_size*y_size; i++ )
		{
		*p++ = (unsigned char) ' ';
		*p++ = csr_col;
		}
	Out->SetAttribute(Out, csr_col);
	Out->ClearScreen(Out);
	csr_x = 0;
	csr_y = 0;
	}
/*...e*/
/*...sth_setcol:0:*/
/* Cleverly TH colours match EFI colours */

void th_setcol(int col)
	{
	csr_col = (unsigned char) (col & 0x7f);
	}
/*...e*/
/*...sth_setcsr:0:*/
void th_setcsr(int csr)
	{
	csr_type = csr;
	}
/*...e*/
/*...sth_inslin:0:*/
void th_inslin(void)
	{
	int i;
	for ( i = y_size-1; i > csr_y; i-- )
		{
		memmove(screen + ( i   *x_size)*2,
		        screen + ((i-1)*x_size)*2,
		        x_size*2);
		restore(0, i, x_size);
		}
	wipe(0, csr_y, x_size);
	csr_x = 0;
	}
/*...e*/
/*...sth_dellin:0:*/
void th_dellin(void)
	{
	int i;
	memmove(screen+( csr_y   *x_size)*2,
	        screen+((csr_y+1)*x_size)*2,
	        (y_size-1-csr_y)*x_size*2);
	for ( i = csr_y; i < y_size-1; i++ )
		restore(0, i, x_size);
	wipe(0, y_size-1, x_size);
	csr_x = 0;
	}
/*...e*/
/*...sth_deleol:0:*/
void th_deleol(void)
	{
	wipe(csr_x, csr_y, x_size-csr_x);
	}
/*...e*/
/*...sth_save:0:*/
void th_save(int x, int y, int n, void *buf)
	{
	unsigned char *p = screen + (y*x_size+x)*2;
	memcpy(buf, p, n*2);
	}
/*...e*/
/*...sth_restore:0:*/
void th_restore(int x, int y, int n, const void *buf)
	{
	unsigned char *p = screen + (y*x_size+x)*2;
	memcpy(p, buf, n*2);
	restore(x, y, n);
	Out->SetAttribute(Out, csr_col);
	}
/*...e*/
/*...sth_isprint:0:*/
/* All characters except NUL, BEL, BS, TAB, LF and CR are printable.
   Maybe more perhaps.*/

TH_BOOLEAN th_isprint(int c)
	{
	return c != 0    &&
	       c != 0x07 &&
	       c != 0x08 &&
	       c != 0x09 &&
	       c != 0x0a &&
	       c != 0x0d;
	}
/*...e*/
/*...sth_p_raw_chr:0:*/
void th_p_raw_chr(register int c)
	{
	unsigned char *p = screen + (csr_y*x_size+csr_x)*2;

	/* Avoid writing to bottom right */
	if ( csr_x != x_size-1 || csr_y != y_size-1 )
		{
		char_buf[0] = (CHAR16) c;
		Out->SetCursorPosition(Out, csr_x, csr_y);
		Out->SetAttribute(Out, csr_col);
		Out->OutputString(Out, char_buf);
		}

	p[0] = (unsigned char) c;
	p[1] = csr_col;
	if ( ++csr_x >= x_size )
		/* moved to next line */
		{
		csr_x -= x_size;
		csr_y++;
		}
	}
/*...e*/
/*...sth_p_raw_str:0:*/
void th_p_raw_str(register const char *s)
	{
	unsigned char *p = screen + (csr_y*x_size+csr_x)*2;
	int len = strlen(s);
	int i;

	for ( i = 0; i < len; i++, s++ )
		{
		*p++ = (unsigned char) *s;
		*p++ = csr_col;
		line_buf[i] = (CHAR16) *s;
		}
	line_buf[i] = (CHAR16) 0;

	/* Avoid writing to bottom right */
	if ( csr_y == y_size-1 && csr_x + len >= x_size && len > 0 )
		line_buf[x_size-csr_x-1] = (CHAR16) 0;

	Out->SetCursorPosition(Out, csr_x, csr_y);
	Out->SetAttribute(Out, csr_col);
	Out->OutputString(Out, line_buf);
	if ( (csr_x += len) >= x_size )
		{
		csr_x -= x_size;
		csr_y++;
		}
	}
/*...e*/
/*...sth_p_chr:0:*/
void th_p_chr(register int c)
	{
	if ( c == '\n' )
		{ csr_x = 0; ++csr_y; }
	else if ( c == '\t' )
		{
		if ( (csr_x = ((csr_x+8) & ~7)) == x_size )
			{
			csr_x = 0; csr_y++;
			}
		}
	else if ( c == '\b' )
		{
		if ( csr_x )
			csr_x--;
		else
			{
			csr_x = x_size - 1;
			csr_y--;
			}
		}
	else if ( c == '\r' )
		csr_x = 0;
	else
		th_p_raw_chr(c);
	}
/*...e*/
/*...sth_p_str:0:*/
void th_p_str(register const char *s)
	{
	while ( *s )
		th_p_chr(*s++);
	}
/*...e*/
/*...sth_sync_screen:0:*/
void th_sync_screen(void)
	{
	if ( csr_y < y_size )
		{
		Out->SetCursorPosition(Out, csr_x, csr_y);
		switch ( csr_type )
			{
			case CSR_OFF:
				Out->EnableCursor(Out, TH_FALSE);
				break;
			case CSR_BLOCK:
			case CSR_ULINE:
			case CSR_ALT:
				Out->EnableCursor(Out, TH_TRUE);
				break;
			}
		}
	else
		Out->EnableCursor(Out, TH_FALSE);
	Out->SetAttribute(Out, csr_col);
	}
/*...e*/
/*...skeyboard stuff:0:*/
typedef struct
	{
	UINT16 ScanCode;
	int k;
	} SCAN;

/* Currently missing from include/efi/eficon.h.
   Values gleaned from efishell/hexedit/hexedit.h. */
#define SCAN_F11 0x15
#define SCAN_F12 0x16

static SCAN scans[] =
	{
		{SCAN_UP,		K_UP},
		{SCAN_DOWN,		K_DOWN},
		{SCAN_RIGHT,		K_RIGHT},
		{SCAN_LEFT,		K_LEFT},
		{SCAN_HOME,		K_HOME},
		{SCAN_END,		K_END},
		{SCAN_INSERT,		K_INSERT},
		{SCAN_DELETE,		K_DELETE},
		{SCAN_PAGE_UP,		K_PGUP},
		{SCAN_PAGE_DOWN,	K_PGDN},
		{SCAN_F1,		K_F1},
		{SCAN_F2,		K_F2},
		{SCAN_F3,		K_F3},
		{SCAN_F4,		K_F4},
		{SCAN_F5,		K_F5},
		{SCAN_F6,		K_F6},
		{SCAN_F7,		K_F7},
		{SCAN_F8,		K_F8},
		{SCAN_F9,		K_F9},
		{SCAN_F10,		K_F10},
		{SCAN_F11,		K_F11},
		{SCAN_F12,		K_F12},
		{SCAN_ESC,		K_ESC},
	};

static int convert_key(EFI_INPUT_KEY Key)
	{
	int i;
	for ( i = 0; i < sizeof(scans)/sizeof(scans[0]); i++ )
		if ( Key.ScanCode == scans[i].ScanCode )
			return scans[i].k;
	if ( Key.ScanCode == SCAN_NULL &&
	     Key.UnicodeChar < 0x100 )
		return Key.UnicodeChar;
	return 0;
	}

static TH_BOOLEAN escprefix(int k)
	{
	return (k >= K_F1 && k <= K_F12) ||
	       (k >= 'a'  && k <= 'z'  ) ||
	       (k >= 'A'  && k <= 'Z'  ) ||
	       (k >= '0'  && k <= '9'  ) ;
	}

static int altify(int k)
	{
	if ( k >= 'a' && k <= 'z' )
		k = toupper(k);
	return k | _K_ALT;
	}

static int keys[2] = { 0, 0 };

static int key_read(UINT64 Timeout)
	{
	EFI_INPUT_KEY Key;
	EFI_STATUS Status;
	WaitForSingleEvent(In->WaitForKey, Timeout);
	Status = In->ReadKeyStroke(In, &Key);
	if ( EFI_ERROR(Status) )
		return FALSE;
	return convert_key(Key);
	}

#define	ESC_DELAY 7000000 /* 7000000*100ns = 0.7s */

int th_key_read(void)
	{
	th_sync_screen();
	while ( keys[0] == 0 )
		keys[0] = key_read(0);
	if ( keys[0] == K_ESC )
		keys[1] = key_read(ESC_DELAY);
	if ( escprefix(keys[1]) )
		{
		int k = altify(keys[1]);
		keys[0] = keys[1] = 0;
		return k;
		}
	else
		{
		int k = keys[0];
		keys[0] = keys[1];
		keys[1] = 0;
		return k;
		}
	}

int th_key_status(void)
	{
	if ( keys[0] == 0 )
		{
		keys[0] = key_read(1);
		if ( keys[0] == K_ESC )
			keys[1] = key_read(ESC_DELAY);
		}
	if ( escprefix(keys[1]) )
		return altify(keys[1]);
	else
		return keys[0];
	}

int th_key_break(void)
	{
	return 0;
	}
/*...e*/
/*...sth_init and th_deinit:0:*/
static TH_BOOLEAN th_inited = 0;
static INT32 init_col;

/*...sget_mode:0:*/
static void get_mode(int *x_size, int *y_size)
	{
	UINTN cols, rows;
	Out->QueryMode(Out, Out->Mode->Mode, &cols, &rows);
	*x_size = (int) cols;
	*y_size = (int) rows;
	}
/*...e*/
/*...sset_mode:0:*/
static int set_mode(int x_size, int y_size)
	{
	int i;
	for ( i = 0; i < Out->Mode->MaxMode; i++ )
		{
		UINTN cols, rows;
		Out->QueryMode(Out, (UINTN) i, &cols, &rows);
		if ( x_size == (int) cols &&
		     y_size == (int) rows )
			{
			Out->SetMode(Out, (UINTN) i);
			return 1;
			}
		}
	return 0;
	}
/*...e*/

TH_BOOLEAN th_init(int th_mode, int desired_x_size, int desired_y_size)
	{
	EFI_STATUS Status;
	int i;

	InitializeLib(_LIBC_EFIImageHandle, _LIBC_EFISystemTable);
	Status = In->Reset(In, /*ExtendedVerification=*/TH_FALSE);
	if ( EFI_ERROR(Status) )
		return TH_FALSE;
	init_col = Out->Mode->Attribute;
	get_mode(&init_x_size, &init_y_size);
	if ( desired_x_size == 0 )
		desired_x_size = init_x_size;
	if ( desired_y_size == 0 )
		desired_y_size = init_y_size;

	if ( desired_x_size > L_BUF )
		return TH_FALSE;

	if ( (screen = (unsigned char *) malloc(desired_x_size*desired_x_size*2)) == NULL )
		return 0;
	if ( desired_x_size == init_x_size &&
	     desired_y_size == init_y_size )
		{
		x_size = init_x_size;
		y_size = init_y_size;
		}
	else if ( set_mode(desired_x_size, desired_y_size) )
		{
		x_size = desired_x_size;
		y_size = desired_y_size;
		}
	else
		{
		free(screen);
		return TH_FALSE;
		}
	th_setcol(COL(COL_WHITE,COL_BLACK,TH_FALSE));
	th_cls();

	char_buf[1] = (CHAR16) 0;
	for ( i = 0; i < L_BUF; i++ )
		wipe_buf[i] = (CHAR16) ' ';
	wipe_buf[i] = (CHAR16) 0;

	return th_inited = TH_TRUE;
	}

void th_deinit(void)
	{
	if ( !th_inited )
		return;
	th_setcol(COL(COL_WHITE,COL_BLACK,TH_FALSE));
	th_cls();
	set_mode(init_x_size, init_y_size);
	free(screen);
	Out->SetAttribute(Out, init_col);
	Out->ClearScreen(Out);
	th_inited = 0;
	}
/*...e*/
