/*

th.c - Terminal handler (OS/2 version)

This code can be compiled 16 or 32 bit.

*/

/*...sincludes:0:*/
#define INCL_DOS
#define	INCL_BASE
#include <os2.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <malloc.h>
#include <memory.h>
#define	TH_C
#include "th.h"

/*...vth\46\h:0:*/
/*...e*/
/*...stypes:0:*/
typedef unsigned  char  byte;
typedef unsigned short  word;
/*...e*/
/*...svars:0:*/
int x_size;				/* Width of screen                   */
int y_size;				/* Height of screen                  */

static int csr_x;			/* Current x in range 0 to y_size-1  */
static int csr_y;			/* Current y in range 0 to x_size-1  */
static byte csr_col;			/* Current colour                    */
static word fill_char;			/* Character used to fill screen     */

static HKBD hkbd = 0;			/* Always seems to be 0 - Why?       */
static HVIO hvio = 0;			/* Always seems to be 0 - Why?       */
/*...e*/

/*...sth_beep:0:*/
void th_beep(void)
	{
	VioWrtTTY("\007", 1, hvio);
	}
/*...e*/
/*...sth_cxy:0:*/
void th_cxy(int x, int y)
	{
	VioSetCurPos(csr_y = y, csr_x = x, hvio);
	}
/*...e*/
/*...sth_cy:0:*/
void th_cy(int y)
	{
	VioSetCurPos(csr_y = y, csr_x = 0, hvio);
	}
/*...e*/
/*...sth_chome:0:*/
void th_chome(void)
	{
	VioSetCurPos(csr_y = 0, csr_x = 0, hvio);
	}
/*...e*/
/*...sth_cls:0:*/
void th_cls(void)
	{
	VioScrollDn(0, 0, y_size - 1, x_size - 1, y_size, (PBYTE) &fill_char, hvio);
	VioSetCurPos(csr_y = 0, csr_x = 0, hvio);
	}
/*...e*/
/*...sth_setcol:0:*/
void th_setcol(int col)
	{
	csr_col = (byte) col;
	fill_char = (col << 8) + ' ';
	}
/*...e*/
/*...sth_setcsr:0:*/
void th_setcsr(int csr)
	{
	VIOCURSORINFO vci;

	VioGetCurType(&vci, hvio);

	switch ( csr )
		{
		case CSR_BLOCK:
			vci.attr   = 0;
			vci.yStart = 0;
			break;
		case CSR_ULINE:
			vci.attr   = 0;
			vci.yStart = vci.cEnd - 2;
			break;
		case CSR_ALT:
			vci.attr   = 0;
			vci.yStart = vci.cEnd / 2;
			break;
		case CSR_OFF:
			vci.attr = -1;	/* Hidden cursor */
			break;
		}
	VioSetCurType(&vci, hvio);
	}
/*...e*/
/*...sth_inslin:0:*/
void th_inslin(void)
	{
	VioScrollDn(csr_y, 0, y_size - 1, x_size - 1, 1, (PBYTE) &fill_char, hvio);
	}
/*...e*/
/*...sth_dellin:0:*/
void th_dellin(void)
	{
	VioScrollUp(csr_y, 0, y_size - 1, x_size - 1, 1, (PBYTE) &fill_char, hvio);
	}
/*...e*/
/*...sth_deleol:0:*/
void th_deleol(void)
	{
	VioScrollUp(csr_y, csr_x, csr_y, x_size - 1, 1, (PBYTE) &fill_char, hvio);
	}
/*...e*/
/*...sth_save:0:*/
void th_save(int x, int y, int n, void *buf)
	{
	USHORT cb = n * sizeof(word);
	VioReadCellStr(buf, &cb, y, x, hvio);
	}
/*...e*/
/*...sth_restore:0:*/
void th_restore(int x, int y, int n, const void *buf)
	{
	VioWrtCellStr((PBYTE) buf, n*sizeof(word), y, x, hvio);
	}
/*...e*/
/*...sth_isprint:0:*/
/* All characters except NUL are printable. */

TH_BOOLEAN th_isprint(int c)
	{
	return c != 0;
	}
/*...e*/
/*...sth_p_raw_chr:0:*/
void th_p_raw_chr(register int c)
	{
	byte b = (byte) c;
	VioWrtCharStrAtt(&b, 1, csr_y, csr_x, &csr_col, hvio);
	if ( ++csr_x >= x_size )
		/* moved to next line */
		{
		csr_x -= x_size;
		csr_y++;
		}
	VioSetCurPos(csr_y, csr_x, hvio);
	}
/*...e*/
/*...sth_p_raw_str:0:*/
void th_p_raw_str(register const char *s)
	{
	int len = strlen(s);
	VioWrtCharStrAtt((CHAR *) s, len, csr_y, csr_x, &csr_col, hvio);
	if ( (csr_x += len) >= x_size )
		/* moved to next line */
		{
		csr_x -= x_size;
		csr_y++;
		}
	VioSetCurPos(csr_y, csr_x, hvio);
	}
/*...e*/
/*...sth_p_chr:0:*/
void th_p_chr(register int c)
	{
	if ( c == '\n' )
		VioSetCurPos(++csr_y, csr_x = 0, hvio);
	else if ( c == '\t' )
		{
		if ( (csr_x = ((csr_x+8) & 0xf8)) == x_size )
			{
			csr_x = 0; csr_y++;
			}
		VioSetCurPos(csr_y, csr_x, hvio);
		}
	else if ( c == '\b' )
		{
		if ( csr_x )
			csr_x--;
		else
			{
			csr_x = x_size - 1;
			csr_y--;
			}
		VioSetCurPos(csr_y, csr_x, hvio);
		}
	else if ( c == '\r' )
		VioSetCurPos(csr_y, csr_x = 0, hvio);
	else
		th_p_raw_chr(c);
	}
/*...e*/
/*...sth_p_str:0:*/
void th_p_str(register const char *s)
	{
	while ( *s )
		th_p_chr(*s++);
	}
/*...e*/
/*...sth_sync_screen:0:*/
void th_sync_screen(void)
	{
	}
/*...e*/
/*...skeyboard stuff:0:*/
static KBDKEYINFO kki;

/*...sth_key_read:0:*/
int th_key_read(void)
	{
	do
		KbdCharIn(&kki, IO_WAIT, hkbd);
	while ( kki.chScan == 0 && kki.chChar == 0 );

	if ( kki.chScan == 0 )
		/* Then was entered via Alt-xxx mechanism */
		return kki.chChar;

	if ( kki.chChar == 0 || kki.chChar == 0xe0 )
		/* Function key */
		return kki.chScan << 8;

	return kki.chChar;
	}
/*...e*/
/*...sth_key_status:0:*/
int th_key_status(void)
	{
	kki.chScan   = 0;
	kki.chChar   = 0;
	kki.fbStatus = 0;
	if ( KbdPeek(&kki, hkbd) != 0 )
		return 0;
	if ( kki.fbStatus == 0 )
		return 0;
	if ( kki.chScan == 0 && kki.chChar == 0 )
		return 0;

	if ( kki.chScan == 0 )
		/* Then was entered via Alt-xxx mechanism */
		return kki.chChar;

	if ( kki.chChar == 0 || kki.chChar == 0xe0 )
		/* Function key */
		return kki.chScan << 8;

	return kki.chChar;
	}
/*...e*/

static TH_BOOLEAN break_occurred = TH_FALSE;

/*...sth_key_break:0:*/
TH_BOOLEAN th_key_break(void)
	{
	if ( !break_occurred )
		return TH_FALSE;
	break_occurred = TH_FALSE;
	return TH_TRUE;
	}
/*...e*/
/*...sctrl_break:0:*/
#ifdef OS2
static void ctrl_break(int x)
#else
static void ctrl_break()
#endif
	{
#ifdef OS2
	x=x; /* Suppress warning */
#endif
	break_occurred = TH_TRUE;
	signal(SIGBREAK, ctrl_break);	/* Set for next time */
	}
/*...e*/
/*...e*/
/*...sth_init and th_deinit:0:*/
static TH_BOOLEAN th_inited = TH_FALSE;

/*...ssaved state:0:*/
static VIOCURSORINFO saved_vci;		/* Cursor setup on startup           */
static VIOMODEINFO saved_vmi = { 8 };	/* Video mode upon startup           */
static byte *saved_buf;			/* Contents of screen on startup     */
static USHORT saved_csr_x, saved_csr_y;	/* Saved cursor position on startup  */
static KBDINFO saved_ki = { 10 };	/* Keyboard mode flags               */
/*...e*/
/*...sset_mode:0:*/
static TH_BOOLEAN set_mode(int x_size, int y_size)
	{
	static VIOMODEINFO vmi = { 8, 1, 4 };

	vmi.col = x_size;
	vmi.row = y_size;
	return ( VioSetMode(&vmi, hvio) == 0 );
	}
/*...e*/

TH_BOOLEAN th_init(int th_mode, int desired_x_size, int desired_y_size)
	{
	th_mode=th_mode; /* Suppress 'unref arg' compiler warning */

/*...ssave current screen:8:*/
{
int w, h, y;

VioGetMode(&saved_vmi, hvio);
w = saved_vmi.col;
h = saved_vmi.row;
VioGetCurPos(&saved_csr_y, &saved_csr_x, hvio);

if ( (saved_buf = (byte *) malloc(w * h * sizeof(word))) == NULL )
	return ( TH_FALSE );
for ( y = 0; y < h; y++ )
	{
	USHORT cb = w * sizeof(word);
	VioReadCellStr(saved_buf + y * w * sizeof(word), &cb, y, 0, hvio);
	}

VioGetCurType(&saved_vci, hvio);	/* Save cursor type for later        */
}
/*...e*/

	x_size = ( desired_x_size != 0 ) ? desired_x_size : (int) saved_vmi.col;
	y_size = ( desired_y_size != 0 ) ? desired_y_size : (int) saved_vmi.row;

	if ( x_size != (int) saved_vmi.col || y_size != (int) saved_vmi.row )
		if ( !set_mode(x_size, y_size) )
			{
			free(saved_buf);
			return TH_FALSE;
			}

	if ( signal(SIGBREAK, SIG_IGN) != SIG_IGN )
		signal(SIGBREAK, ctrl_break);

	th_setcsr(CSR_BLOCK);
	th_setcol(COL(COL_WHITE, COL_BLACK, TH_FALSE, TH_FALSE));
	th_cls();

/*...sset terminal mode:8:*/
{
KBDINFO ki;

KbdGetStatus(&saved_ki, hkbd);	/* Get keyboard status               */
ki = saved_ki;			/* Keep for restoring later          */
ki.fsMask &= ~(8|1);		/* Remove ASCII bit and Echo On bit  */
ki.fsMask |= (4|2);		/* Turn on binary bit and Echo off   */
KbdSetStatus(&ki, hkbd);	/* Set new mode for keyboard input   */
}
/*...e*/

	return th_inited = TH_TRUE;
	}

void th_deinit(void)
	{
	if ( !th_inited )
		return;

/*...srestore current screen:8:*/
{
int w = saved_vmi.col;
int h = saved_vmi.row;
int y;

VioSetCurType(&saved_vci, hvio);	/* Restore original cursor type      */

if ( x_size != w || y_size != h )
	VioSetMode(&saved_vmi, hvio);	/* Return original screen mode       */

for ( y = 0; y < h; y++ )
	VioWrtCellStr(saved_buf + y * w * 2, w * 2, y, 0, hvio);
free((char *) saved_buf);
VioSetCurPos(saved_csr_y, saved_csr_x, hvio);

KbdSetStatus(&saved_ki, hkbd);	/* Restore old keyboard status       */
}
/*...e*/

	signal(SIGBREAK, SIG_DFL);

	th_inited = TH_FALSE;
	}
/*...e*/
