/*

th.c - Terminal handler (Win32)

*/

/*...sincludes:0:*/
#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <string.h>
#include <malloc.h>
#include <memory.h>
#define	TH_C
#include "th.h"

/*...vth\46\h:0:*/
/*...e*/

/*...svars:0:*/
int x_size;				/* Width of screen                   */
int y_size;				/* Height of screen                  */

static HANDLE hConOut;			/* Handle to console                 */
static HANDLE hConIn;			/* Handle to console                 */
static SMALL_RECT srScreen;		/* The screen                        */
static CHAR_INFO chi;			/* Includes background colour        */

static int csr_x;			/* Current x in range 0 to y_size-1  */
static int csr_y;			/* Current y in range 0 to x_size-1  */

static TH_BOOLEAN buggy95;
/*...e*/

/*...sMyWriteConsole:0:*/
static void MyWriteConsole(HANDLE h, CONST VOID *b, DWORD n, DWORD *m, VOID *pv)
	{
	if ( buggy95 && csr_y == y_size-1 && csr_x + (int) n >= x_size )
		/* Would cause a screen scroll due to bug in Windows 95.
		   We have already disabled ENABLE_WRAP_AT_EOL_OUTPUT, so
		   this shouldn't happen (and doesn't on Windows NT).
		   Temporarily disable ENABLE_PROCESSED_OUTPUT, which
		   prevents the scroll, at the expense of slowing down this
		   particular screen update. */
		{
		DWORD dwMode;
		GetConsoleMode(hConOut, &dwMode);
		SetConsoleMode(hConOut, dwMode & ~ENABLE_PROCESSED_OUTPUT);
		WriteConsole(h, b, n, m, pv);
		SetConsoleMode(hConOut, dwMode);
		}
	else
		WriteConsole(h, b, n, m, pv);
	}
/*...e*/
/*...sgotoxy:0:*/
static void gotoxy(int x, int y)
	{
	COORD coord;
	coord.X = x;
	coord.Y = y;
	SetConsoleCursorPosition(hConOut, coord);
	}
/*...e*/
/*...swipe:0:*/
/* I would have liked to have acheived the clearing of rectangles
   using the ScrollConsoleScreenBuffer, but this doesn't work under
   Windows 95, so I've coded a slower alternative. */

static void wipe(int x, int y, int len)
	{
	COORD coord;
	DWORD dwWritten;
	coord.X = x;
	coord.Y = y;
	FillConsoleOutputCharacter(hConOut, ' ',
		len, coord, &dwWritten);
	FillConsoleOutputAttribute(hConOut, chi.Attributes,
		len, coord, &dwWritten);
	}
/*...e*/
/*...sth_beep:0:*/
void th_beep(void)
	{
	Beep(100,50); /* Frequency=100Hz, duration=50ms */
	}
/*...e*/
/*...sth_cxy:0:*/
void th_cxy(int x, int y)
	{
	csr_x = x;
	csr_y = y;
	}
/*...e*/
/*...sth_cy:0:*/
void th_cy(int y)
	{
	csr_x = 0;
	csr_y = y;
	}
/*...e*/
/*...sth_chome:0:*/
void th_chome(void)
	{
	csr_x = csr_y = 0;
	}
/*...e*/
/*...sth_cls:0:*/
/* Under Windows 95, scrolling a rectangle too far is like the 1 line high
   case, and doesn't appear to work either, so I'll code in a work-around */

void th_cls(void)
	{
#if 0
	COORD coord;
	coord.X = 0;
	coord.Y = y_size;
	ScrollConsoleScreenBuffer(hConOut, &srScreen, &srScreen, coord, &chi);
#else
	int i;
	for ( i = 0; i < y_size; i++ )
		wipe(0, i, x_size);
#endif
	csr_x = csr_y = 0;
	}
/*...e*/
/*...sth_setcol:0:*/
void th_setcol(int col)
	{
	chi.Attributes = 0;
	if ( col & 0x01 ) chi.Attributes |= FOREGROUND_BLUE;
	if ( col & 0x02 ) chi.Attributes |= FOREGROUND_GREEN;
	if ( col & 0x04 ) chi.Attributes |= FOREGROUND_RED;
	if ( col & 0x08 ) chi.Attributes |= FOREGROUND_INTENSITY;
	if ( col & 0x10 ) chi.Attributes |= BACKGROUND_BLUE;
	if ( col & 0x20 ) chi.Attributes |= BACKGROUND_GREEN;
	if ( col & 0x40 ) chi.Attributes |= BACKGROUND_RED;
	if ( col & 0x80 ) chi.Attributes |= BACKGROUND_INTENSITY;
	SetConsoleTextAttribute(hConOut, chi.Attributes);
	}
/*...e*/
/*...sth_setcsr:0:*/
/* Under Windows 95, setting a cursor 100% high doesn't work.
   It vanishes altogether, so I shall ask for a 99% high cursor. */

void th_setcsr(int csr)
	{
	CONSOLE_CURSOR_INFO cci;
	GetConsoleCursorInfo(hConOut, &cci);
	switch ( csr )
		{
		case CSR_OFF:
			cci.bVisible = TH_FALSE;
			break;
		case CSR_BLOCK:
			cci.bVisible = TH_TRUE;
			cci.dwSize   = 99;
			break;
		case CSR_ULINE:
			cci.bVisible = TH_TRUE;
			cci.dwSize   = 20;
			break;
		case CSR_ALT:
			cci.bVisible = TH_TRUE;
			cci.dwSize   = 50;
			break;
		}
	SetConsoleCursorInfo(hConOut, &cci);
	}
/*...e*/
/*...sth_inslin:0:*/
/* Under Windows 95, scrolling a rectangle 1 line high doesn't
   appear to work, so I'll code in a work-around */

void th_inslin(void)
	{
	if ( csr_y < y_size - 1 )
		{
		SMALL_RECT sr;
		COORD coord;
		sr.Left   = 0;
		sr.Top    = csr_y;
		sr.Right  = x_size -1;
		sr.Bottom = y_size -1;
		coord.X = 0;
		coord.Y = csr_y + 1;
		ScrollConsoleScreenBuffer(hConOut, &sr, &srScreen, coord, &chi);
		}
	else
		wipe(0, csr_y, x_size);
	csr_x = 0;
	}
/*...e*/
/*...sth_dellin:0:*/
/* Under Windows 95, scrolling a rectangle 1 line high doesn't
   appear to work, so I'll code in a work-around */

void th_dellin(void)
	{
	if ( csr_y < y_size - 1 )
		{
		SMALL_RECT sr;
		COORD coord;
		sr.Left   = 0;
		sr.Top    = csr_y+1;
		sr.Right  = x_size -1;
		sr.Bottom = y_size -1;
		coord.X = 0;
		coord.Y = csr_y;
		ScrollConsoleScreenBuffer(hConOut, &sr, &srScreen, coord, &chi);
		}
	else
		wipe(0, csr_y, x_size);
	csr_x = 0;
	}
/*...e*/
/*...sth_deleol:0:*/
void th_deleol(void)
	{
	wipe(csr_x, csr_y, x_size-csr_x);
	}
/*...e*/
/*...sth_save:0:*/
void th_save(int x, int y, int n, void *buf)
	{
	COORD coord;
	DWORD dwRead;
	coord.X = (SHORT) x; coord.Y = (SHORT) y;
	ReadConsoleOutputCharacter(hConOut,
		buf, n, coord, &dwRead);
	ReadConsoleOutputAttribute(hConOut,
		(WORD*) ((char *)buf+n), n, coord, &dwRead);
	}
/*...e*/
/*...sth_restore:0:*/
void th_restore(int x, int y, int n, const void *buf)
	{
	COORD coord;
	DWORD dwWritten;
	coord.X = (SHORT) x; coord.Y = (SHORT) y;
	WriteConsoleOutputCharacter(hConOut,
		buf, n, coord, &dwWritten);
	WriteConsoleOutputAttribute(hConOut,
		(WORD*) ((char*)buf+n), n, coord, &dwWritten);
	}
/*...e*/
/*...sth_isprint:0:*/
/* All characters except NUL, BEL, BS, TAB, LF and CR are printable. */

TH_BOOLEAN th_isprint(int c)
	{
	return c != 0    &&
	       c != 0x07 &&
	       c != 0x08 &&
	       c != 0x09 &&
	       c != 0x0a &&
	       c != 0x0d;
	}
/*...e*/
/*...sth_p_raw_chr:0:*/
void th_p_raw_chr(register int c)
	{
	BYTE b = (BYTE) c;
	DWORD dwWritten;
	gotoxy(csr_x, csr_y);
	MyWriteConsole(hConOut, &b, 1, &dwWritten, NULL);
	if ( ++csr_x >= x_size )
		/* moved to next line */
		{
		csr_x -= x_size;
		csr_y++;
		}
	}
/*...e*/
/*...sth_p_raw_str:0:*/
void th_p_raw_str(register const char *s)
	{
	DWORD dwWrite = (DWORD) strlen(s);
	DWORD dwWritten;
	gotoxy(csr_x, csr_y);
	MyWriteConsole(hConOut, s, dwWrite, &dwWritten, NULL);
	if ( (csr_x += dwWrite) >= x_size )
		/* moved to next line */
		{
		csr_x -= x_size;
		csr_y++;
		}
	}
/*...e*/
/*...sth_p_chr:0:*/
void th_p_chr(register int c)
	{
	if ( c == '\n' )
		{ csr_x = 0; ++csr_y; }
	else if ( c == '\t' )
		{
		if ( (csr_x = ((csr_x+8) & 0xf8)) == x_size )
			{
			csr_x = 0; csr_y++;
			}
		}
	else if ( c == '\b' )
		{
		if ( csr_x )
			csr_x--;
		else
			{
			csr_x = x_size - 1;
			csr_y--;
			}
		}
	else if ( c == '\r' )
		csr_x = 0;
	else
		th_p_raw_chr(c);
	}
/*...e*/
/*...sth_p_str:0:*/
void th_p_str(register const char *s)
	{
	while ( *s )
		th_p_chr(*s++);
	}
/*...e*/
/*...sth_sync_screen:0:*/
void th_sync_screen(void)
	{
	gotoxy(csr_x, csr_y);
	}
/*...e*/
/*...skeyboard stuff:0:*/
/* Only interested in keyboard events.
   Only interested in key presses, not releases.
   Combinations of Shift+Ctrl+Alt are not allowed.
   The Shift, Ctrl, Alt and Caps Lock keys on their own are not allowed. */

static INPUT_RECORD ir;
static WORD wRepeatCount = 0;

#define	ALT_PRESSED  LEFT_ALT_PRESSED
#define	CTRL_PRESSED (LEFT_CTRL_PRESSED|RIGHT_CTRL_PRESSED)

/*...sgood_event:0:*/
static TH_BOOLEAN good_event(INPUT_RECORD *ir)
	{
	DWORD dwMeta;

	if ( ir->EventType != KEY_EVENT )
		return TH_FALSE;

	if ( !ir->Event.KeyEvent.bKeyDown )
		return TH_FALSE;

	dwMeta = ir->Event.KeyEvent.dwControlKeyState;
	if ( ( (dwMeta & ALT_PRESSED  ) != 0 ? 1 : 0 ) +
	     ( (dwMeta & CTRL_PRESSED ) != 0 ? 1 : 0 ) +
	     ( (dwMeta & SHIFT_PRESSED) != 0 ? 1 : 0 ) > 1 )
		return TH_FALSE;

	if ( ir->Event.KeyEvent.wVirtualKeyCode == VK_SHIFT   ||
	     ir->Event.KeyEvent.wVirtualKeyCode == VK_CONTROL ||
	     ir->Event.KeyEvent.wVirtualKeyCode == VK_MENU    ||  /* = Alt */
	     ir->Event.KeyEvent.wVirtualKeyCode == VK_CAPITAL )
		return TH_FALSE;

	return TH_TRUE;
	}
/*...e*/
/*...sdecode_key:0:*/
static int decode_key(INPUT_RECORD *ir)
	{
	DWORD dwMeta = ir->Event.KeyEvent.dwControlKeyState;
	int k = ((ir->Event.KeyEvent.uChar.AsciiChar) & 0xff);
	if ( k != 0 )
		{
		if ( dwMeta & ALT_PRESSED )
			{
			if ( islower(k) )
				k = toupper(k);
			k |= _K_ALT;
			}
		return k;
		}

	k = ir->Event.KeyEvent.wVirtualKeyCode << 8;
	if ( dwMeta & SHIFT_PRESSED ) k |= _K_SHIFT;
	if ( dwMeta & CTRL_PRESSED  ) k |= _K_CTRL ;
	if ( dwMeta & ALT_PRESSED   ) k |= _K_ALT  ;
	return k;
	}
/*...e*/

int th_key_read(void)
	{
	th_sync_screen();
	if ( wRepeatCount == 0 )
		{
		DWORD dwActual;
		do
			ReadConsoleInput(hConIn, &ir, 1, &dwActual);
		while ( !good_event(&ir) );
		wRepeatCount = ir.Event.KeyEvent.wRepeatCount;
		}

	--wRepeatCount;

	return decode_key(&ir);
	}

int th_key_status(void)
	{
	th_sync_screen();
	if ( wRepeatCount > 0 )
		return decode_key(&ir);
	for ( ;; )
		{
		DWORD dwActual;
		PeekConsoleInput(hConIn, &ir, 1, &dwActual);
		if ( dwActual < 1 )
			return 0;
		if ( good_event(&ir) )
			return decode_key(&ir);
		ReadConsoleInput(hConIn, &ir, 1, &dwActual);
		}
	/* Can't get here */
	}

static TH_BOOLEAN break_occurred = TH_FALSE;

TH_BOOLEAN th_key_break(void)
	{
	TH_BOOLEAN occurred = break_occurred;
	break_occurred = TH_FALSE;
	return occurred;
	}
/*...e*/
/*...sth_init and th_deinit:0:*/
static TH_BOOLEAN th_inited = TH_FALSE;

static CONSOLE_CURSOR_INFO cciSaved;
static CONSOLE_SCREEN_BUFFER_INFO csbiSaved;
static BYTE *pbCharsSaved;
static WORD *pwAttrsSaved;
static DWORD dwModeSavedIn, dwModeSavedOut;

/*...sctrl_break:0:*/
static void ctrl_break(int x)
	{
	x=x; /* Suppress compiler warning */
	break_occurred = TH_TRUE;
	signal(SIGBREAK, ctrl_break);	/* Set for next time */
	}
/*...e*/
/*...sset_mode:0:*/
/* Change the screen size.
   Must consider the effect of the buffer and window size.
   The 'window' always shows a subset of the 'buffer'. */

static TH_BOOLEAN set_mode(int nx, int ny)
	{
	CONSOLE_SCREEN_BUFFER_INFO csbi;
	SMALL_RECT sr;
	COORD coord;
	int ox, oy, mx, my;

	GetConsoleScreenBufferInfo(hConOut, &csbi);
	ox = csbi.dwSize.X;
	oy = csbi.dwSize.Y;

	sr.Left   = 0;
	sr.Right  = min(ox,nx) - 1;
	sr.Top    = 0;
	sr.Bottom = min(oy,ny) - 1;
	if ( !SetConsoleWindowInfo(hConOut, TH_TRUE, &sr) )
		return TH_FALSE;

	coord.X = nx;
	coord.Y = ny;
	if ( !SetConsoleScreenBufferSize(hConOut, coord) )
		return TH_FALSE;

	GetConsoleScreenBufferInfo(hConOut, &csbi);
	mx = csbi.dwMaximumWindowSize.X;
	my = csbi.dwMaximumWindowSize.Y;

	sr.Left   = 0;
	sr.Right  = min(mx,nx) - 1;
	sr.Top    = 0;
	sr.Bottom = min(my,ny) - 1;
	SetConsoleWindowInfo(hConOut, TH_TRUE, &sr);

	return TH_TRUE;
	}
/*...e*/

TH_BOOLEAN th_init(int th_mode, int desired_x_size, int desired_y_size)
	{
	int i, x_size_init, y_size_init;
	OSVERSIONINFO osvi;

	th_mode=th_mode; /* Suppress 'unref arg' compiler warning */

	hConIn  = GetStdHandle(STD_INPUT_HANDLE);
	hConOut = GetStdHandle(STD_OUTPUT_HANDLE);

	GetConsoleScreenBufferInfo(hConOut, &csbiSaved);
	GetConsoleCursorInfo(hConOut, &cciSaved);
	x_size_init = csbiSaved.dwSize.X;
	y_size_init = csbiSaved.dwSize.Y;
	if ( (pbCharsSaved = (BYTE *) malloc(x_size_init * y_size_init * sizeof(BYTE))) == NULL )
		return TH_FALSE;
	if ( (pwAttrsSaved = (WORD *) malloc(x_size_init * y_size_init * sizeof(WORD))) == NULL )
		{
		free(pbCharsSaved);
		return TH_FALSE;
		}
	for ( i = 0; i < y_size_init; i++ )
		{
		COORD coord;
		DWORD dwRead;
		coord.X = (SHORT) 0; coord.Y = (SHORT) i;
		ReadConsoleOutputCharacter(hConOut,
			pbCharsSaved + i * x_size_init,
			x_size_init, coord, &dwRead);
		ReadConsoleOutputAttribute(hConOut,
			pwAttrsSaved + i * x_size_init,
			x_size_init, coord, &dwRead);
		}

	x_size = ( desired_x_size != 0 ) ? desired_x_size : x_size_init;
	y_size = ( desired_y_size != 0 ) ? desired_y_size : y_size_init;

	if ( x_size != x_size_init || y_size != y_size_init )
		if ( !set_mode(x_size, y_size) )
			{
			free(pwAttrsSaved);
			free(pbCharsSaved);
			return TH_FALSE;
			}

	if ( signal(SIGBREAK, SIG_IGN) != SIG_IGN )
		signal(SIGBREAK, ctrl_break);

	srScreen.Left   = 0;
	srScreen.Top    = 0;
	srScreen.Right  = x_size -1;
	srScreen.Bottom = y_size -1;
	chi.Char.AsciiChar = ' ';

	th_setcsr(CSR_BLOCK);
	th_setcol(COL(COL_WHITE, COL_BLACK, TH_FALSE, TH_FALSE));
	th_cls();

	GetConsoleMode(hConIn, &dwModeSavedIn);
	SetConsoleMode(hConIn, 0);

	/* We need to run with ENABLE_WRAP_AT_EOL_OUTPUT disabled,
	   so that writing to the bottom right character doesn't
	   automatically scroll the screen. However this doesn't work
	   on Windows 95. Disabling ENABLE_PROCESSED_OUTPUT fixes
	   the problem at a severe performance penalty. I only disable
	   this flag when updating the bottom line of the screen. */

	GetConsoleMode(hConOut, &dwModeSavedOut);
	SetConsoleMode(hConOut, ENABLE_PROCESSED_OUTPUT);

	osvi.dwOSVersionInfoSize = sizeof(osvi);
	GetVersionEx(&osvi);
	buggy95 = ( osvi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS );

	return th_inited = TH_TRUE;
	}

void th_deinit(void)
	{
	int i, x_size_init, y_size_init;

	if ( !th_inited )
		return;

	x_size_init = csbiSaved.dwSize.X;
	y_size_init = csbiSaved.dwSize.Y;
	if ( x_size != x_size_init || y_size != y_size_init )
		set_mode(x_size_init, y_size_init);

	for ( i = 0; i < y_size_init; i++ )
		{
		COORD coord;
		DWORD dwWritten;
		coord.X = (SHORT) 0; coord.Y = (SHORT) i;
		WriteConsoleOutputCharacter(hConOut,
			pbCharsSaved + i * x_size_init,
			x_size_init, coord, &dwWritten);
		WriteConsoleOutputAttribute(hConOut,
			pwAttrsSaved + i * x_size_init,
			x_size_init, coord, &dwWritten);
		}
	free(pwAttrsSaved);
	free(pbCharsSaved);

	gotoxy(csbiSaved.dwCursorPosition.X, csbiSaved.dwCursorPosition.Y);
	SetConsoleTextAttribute(hConOut, csbiSaved.wAttributes);

	SetConsoleMode(hConIn , dwModeSavedIn );
	SetConsoleMode(hConOut, dwModeSavedOut);
	SetConsoleCursorInfo(hConOut, &cciSaved);

	signal(SIGBREAK, SIG_DFL);

	th_inited = TH_FALSE;
	}
/*...e*/
