/*

th.h - Interface to Terminal Handler (Win32 version)

*/

#ifndef TH_H
#define	TH_H

#define	WIN32_LEAN_AND_MEAN
#include <windows.h>
#undef WIN32_LEAN_AND_MEAN

typedef int TH_BOOLEAN;
#define	TH_TRUE  1
#define	TH_FALSE 0

#define	CSR_BLOCK	0
#define	CSR_OFF		1
#define	CSR_ULINE       2
#define	CSR_ALT		3

#define	COL_BLACK	0
#define	COL_BLUE	1
#define	COL_GREEN	2
#define	COL_CYAN	3
#define	COL_RED		4
#define	COL_MAGENTA	5
#define	COL_YELLOW	6
#define	COL_WHITE	7

#define	COL(f,b,fb,bb)	((f)+(b)*0x10+(fb)*0x08+(bb)*0x80)

#define	MODE_NORMAL	0x0000

#define	K_CTRL(c)	((c)&0x1f)
#define	K_BS		0x0008
#define	K_TAB		0x0009
#define	K_CR		0x000d
#define	K_LF		0x000a
#define	K_ESC		0x001b
#define	K_DEL		0x007f

#define	_K_SHIFT        0x10000
#define	_K_CTRL         0x20000
#define _K_ALT          0x40000

#define	K_F1		(VK_F1<<8)
#define	K_F2		(VK_F2<<8)
#define	K_F3		(VK_F3<<8)
#define	K_F4		(VK_F4<<8)
#define	K_F5		(VK_F5<<8)
#define	K_F6		(VK_F6<<8)
#define	K_F7		(VK_F7<<8)
#define	K_F8		(VK_F8<<8)
#define	K_F9		(VK_F9<<8)
#define	K_F10		(VK_F10<<8)
#define	K_F11		(VK_F11<<8)
#define	K_F12		(VK_F12<<8)

#define	K_SHIFT_F1	(_K_SHIFT|K_F1)
#define	K_SHIFT_F2	(_K_SHIFT|K_F2)
#define	K_SHIFT_F3	(_K_SHIFT|K_F3)
#define	K_SHIFT_F4	(_K_SHIFT|K_F4)
#define	K_SHIFT_F5	(_K_SHIFT|K_F5)
#define	K_SHIFT_F6	(_K_SHIFT|K_F6)
#define	K_SHIFT_F7	(_K_SHIFT|K_F7)
#define	K_SHIFT_F8	(_K_SHIFT|K_F8)
#define	K_SHIFT_F9	(_K_SHIFT|K_F9)
#define	K_SHIFT_F10	(_K_SHIFT|K_F10)
#define	K_SHIFT_F11	(_K_SHIFT|K_F11)
#define	K_SHIFT_F12	(_K_SHIFT|K_F12)

#define	K_CTRL_F1	(_K_CTRL|K_F1)
#define	K_CTRL_F2	(_K_CTRL|K_F2)
#define	K_CTRL_F3	(_K_CTRL|K_F3)
#define	K_CTRL_F4	(_K_CTRL|K_F4)
#define	K_CTRL_F5	(_K_CTRL|K_F5)
#define	K_CTRL_F6	(_K_CTRL|K_F6)
#define	K_CTRL_F7	(_K_CTRL|K_F7)
#define	K_CTRL_F8	(_K_CTRL|K_F8)
#define	K_CTRL_F9	(_K_CTRL|K_F9)
#define	K_CTRL_F10	(_K_CTRL|K_F10)
#define	K_CTRL_F11	(_K_CTRL|K_F11)
#define	K_CTRL_F12	(_K_CTRL|K_F12)

#define	K_ALT_F1	(_K_ALT|K_F1)
#define	K_ALT_F2	(_K_ALT|K_F2)
#define	K_ALT_F3	(_K_ALT|K_F3)
#define	K_ALT_F4	(_K_ALT|K_F4)
#define	K_ALT_F5	(_K_ALT|K_F5)
#define	K_ALT_F6	(_K_ALT|K_F6)
#define	K_ALT_F7	(_K_ALT|K_F7)
#define	K_ALT_F8	(_K_ALT|K_F8)
#define	K_ALT_F9	(_K_ALT|K_F9)
#define	K_ALT_F10	(_K_ALT|K_F10)
#define	K_ALT_F11	(_K_ALT|K_F11)
#define	K_ALT_F12	(_K_ALT|K_F12)

#define	K_UP		(VK_UP<<8)
#define	K_DOWN		(VK_DOWN<<8)
#define	K_LEFT		(VK_LEFT<<8)
#define	K_RIGHT		(VK_RIGHT<<8)
#define	K_PGUP		(VK_PRIOR<<8)
#define	K_PGDN		(VK_NEXT<<8)
#define	K_INSERT	(VK_INSERT<<8)
#define	K_DELETE	(VK_DELETE<<8)
#define	K_HOME		(VK_HOME<<8)
#define	K_END		(VK_END<<8)

#define	K_CTRL_UP	(_K_CTRL|K_UP)
#define	K_CTRL_DOWN	(_K_CTRL|K_DOWN)
#define	K_CTRL_LEFT	(_K_CTRL|K_LEFT)
#define	K_CTRL_RIGHT	(_K_CTRL|K_RIGHT)
#define	K_CTRL_PGUP	(_K_CTRL|K_PGUP)
#define	K_CTRL_PGDN	(_K_CTRL|K_PGDN)
#define	K_CTRL_INSERT	(_K_CTRL|K_INSERT)
#define	K_CTRL_DELETE	(_K_CTRL|K_DELETE)
#define	K_CTRL_HOME	(_K_CTRL|K_HOME)
#define	K_CTRL_END	(_K_CTRL|K_END)

#define	K_ALT_UP	(_K_ALT|K_UP)
#define	K_ALT_DOWN	(_K_ALT|K_DOWN)
#define	K_ALT_LEFT	(_K_ALT|K_LEFT)
#define	K_ALT_RIGHT	(_K_ALT|K_RIGHT)
#define	K_ALT_PGUP	(_K_ALT|K_PGUP)
#define	K_ALT_PGDN	(_K_ALT|K_PGDN)
#define	K_ALT_INSERT	(_K_ALT|K_INSERT)
#define	K_ALT_DELETE	(_K_ALT|K_DELETE)
#define	K_ALT_HOME	(_K_ALT|K_HOME)
#define	K_ALT_END	(_K_ALT|K_END)

#define	K_ALT_A		(_K_ALT|'A')
#define	K_ALT_B		(_K_ALT|'B')
#define	K_ALT_C		(_K_ALT|'C')
#define	K_ALT_D		(_K_ALT|'D')
#define	K_ALT_E		(_K_ALT|'E')
#define	K_ALT_F		(_K_ALT|'F')
#define	K_ALT_G		(_K_ALT|'G')
#define	K_ALT_H		(_K_ALT|'H')
#define	K_ALT_I		(_K_ALT|'I')
#define	K_ALT_J		(_K_ALT|'J')
#define	K_ALT_K		(_K_ALT|'K')
#define	K_ALT_L		(_K_ALT|'L')
#define	K_ALT_M		(_K_ALT|'M')
#define	K_ALT_N		(_K_ALT|'N')
#define	K_ALT_O		(_K_ALT|'O')
#define	K_ALT_P		(_K_ALT|'P')
#define	K_ALT_Q		(_K_ALT|'Q')
#define	K_ALT_R		(_K_ALT|'R')
#define	K_ALT_S		(_K_ALT|'S')
#define	K_ALT_T		(_K_ALT|'T')
#define	K_ALT_U		(_K_ALT|'U')
#define	K_ALT_V		(_K_ALT|'V')
#define	K_ALT_W		(_K_ALT|'W')
#define	K_ALT_X		(_K_ALT|'X')
#define	K_ALT_Y		(_K_ALT|'Y')
#define	K_ALT_Z		(_K_ALT|'Z')

#define	K_ALT_0		(_K_ALT|'0')
#define	K_ALT_1		(_K_ALT|'1')
#define	K_ALT_2		(_K_ALT|'2')
#define	K_ALT_3		(_K_ALT|'3')
#define	K_ALT_4		(_K_ALT|'4')
#define	K_ALT_5		(_K_ALT|'5')
#define	K_ALT_6		(_K_ALT|'6')
#define	K_ALT_7		(_K_ALT|'7')
#define	K_ALT_8		(_K_ALT|'8')
#define	K_ALT_9		(_K_ALT|'9')

#define	K_5		(VK_MBUTTON<<8)

#define	TH_SAVE_RESTORE_CELL 3

#ifndef TH_C

extern int x_size, y_size;

extern TH_BOOLEAN th_init(int mode, int x_size, int y_size);
extern void th_deinit(void);
extern void th_beep(void);
extern void th_cls(void);
extern void th_cxy(int x, int y);
extern void th_cy(int y);
extern void th_chome(void);
extern void th_setcol(int col);
extern void th_setcsr(int csr);
extern void th_inslin(void);
extern void th_dellin(void);
extern void th_deleol(void);
extern void th_save(int x, int y, int n, void *buf);
extern void th_restore(int x, int y, int n, const void *buf);
extern TH_BOOLEAN th_isprint(int c);
extern void th_p_raw_chr(int c);
extern void th_p_raw_str(const char *s);
extern void th_p_chr(int c);
extern void th_p_str(const char *s);
extern void th_sync_screen(void);

extern int th_key_read(void);
extern int th_key_status(void);
extern TH_BOOLEAN th_key_break(void);

#endif

#endif
